from typing import Dict, List, Optional, Tuple, Union

from emoji import UNICODE_EMOJI
from spacy.language import Language
from spacy.matcher import PhraseMatcher
from spacy.tokens import Doc, Span, Token
from spacy.util import filter_spans

# Make sure multi-character emoji don't contain whitespace
EMOJI = {e.replace(" ", ""): t for e, t in UNICODE_EMOJI.items()}

DEFAULT_ATTRS = ("has_emoji", "is_emoji", "emoji_desc", "emoji")

DEFAULT_CONFIG = {}


@Language.factory("twiiter", default_config=DEFAULT_CONFIG)
def create_twitter(
    nlp: Language,
    name: str,
    merge_spans: bool = True,
    lookup: Optional[Dict[str, str]] = None,
    pattern_id: str = "EMOJI",
    attrs: Tuple[str, str, str, str] = DEFAULT_ATTRS,
    force_extension: bool = True,
):
    return Twitter(nlp, merge_spans, lookup, pattern_id, attrs, force_extension)


class Twitter:
    """spaCy pipeline component for added hashtags, mentions and urls to
    `Doc`objects.

    Examples:
        >>> import spacy
        >>> from dacy.twitter import Twitter
        >>> nlp = spacy.load("en_core_web_sm")
        >>> nlp.add_pipe("twitter", first=True)
        >>> doc = nlp("@Charles life is great right www.greatlife.com! #yolo")
        >>> doc._.has_hashtag
        True
        >>> doc[2:5]._.has_hashtag
        True
        >>> doc[-1]._.is_hashtag
        True
        >>> doc[0]._.is_mention is True
        True
        >>> doc._.hashtags
        ["#yolo"]
        >>> doc._.hastag_indexes
        [(48, 53)]
        >>> doc[0]._.is_mention
        True
        >>> doc[6]._.is_url
        True
    """

    name = "twitter"

    def __init__(
        self,
        nlp: Language,
        redo_tokenization: bool = True,
        lookup: Optional[Dict[str, str]] = None,
        pattern_id: str = "EMOJI",
        attrs: Tuple[str, str, str, str] = DEFAULT_ATTRS,
        force_extension: bool = True,
    ) -> None:
        """Initialise the pipeline component.

        nlp (Language): The shared nlp object. Used to initialise the
        matcher     with the shared `Vocab`, and create `Doc` match
        patterns. attrs (tuple): Attributes to set on the ._ property.
        Defaults to     ('has_emoji', 'is_emoji', 'emoji_desc',
        'emoji'). pattern_id (unicode): ID of match pattern, defaults to
        'EMOJI'. Can be     changed to avoid ID clashes. merge_spans
        (bool): Merge spans containing multi-character emoji. Will
        only merge combined emoji resulting in one icon, not sequences.
        lookup (dict): Optional lookup table that maps emoji unicode
        strings     to custom descriptions, e.g. translations or other
        annotations. RETURNS (callable): A spaCy pipeline component.
        """
        self._has_emoji, self._is_emoji, self._emoji_desc, self._emoji = attrs
        self.merge_spans = merge_spans
        self.lookup = lookup or {}
        self.matcher = PhraseMatcher(nlp.vocab)
        emoji_patterns = list(nlp.tokenizer.pipe(EMOJI.keys()))
        self.matcher.add(pattern_id, None, *emoji_patterns)
        # Add attributes
        Doc.set_extension(self._has_emoji, getter=self.has_emoji, force=force_extension)
        Doc.set_extension(self._emoji, getter=self.iter_emoji, force=force_extension)
        Span.set_extension(
            self._has_emoji,
            getter=self.has_emoji,
            force=force_extension,
        )
        Span.set_extension(self._emoji, getter=self.iter_emoji, force=force_extension)
        Token.set_extension(self._is_emoji, default=False, force=force_extension)
        Token.set_extension(
            self._emoji_desc,
            getter=self.get_emoji_desc,
            force=force_extension,
        )

    def __call__(self, doc: Doc) -> Doc:
        """Apply the pipeline component to a `Doc` object.

        doc (Doc): The `Doc` returned by the previous pipeline
        component. RETURNS (Doc): The modified `Doc` object.
        """
        spans = self.matcher(doc, as_spans=True)
        for span in spans:
            for token in span:
                token._.set(self._is_emoji, True)

        if self.merge_spans:
            spans = filter_spans(spans)
            with doc.retokenize() as retokenizer:
                for span in spans:
                    if len(span) > 1:
                        retokenizer.merge(span)
        return doc

    def has_emoji(self, tokens: Union[Doc, Span]) -> bool:
        return any(token._.get(self._is_emoji) for token in tokens)

    def iter_emoji(self, tokens: Union[Doc, Span]) -> List[Tuple[str, int, str]]:
        return [
            (t.text, i, t._.get(self._emoji_desc))
            for i, t in enumerate(tokens)
            if t._.get(self._is_emoji)
        ]

    def get_emoji_desc(self, token: Token) -> Optional[str]:
        if token.text in self.lookup:
            return self.lookup[token.text]
        if token.text in EMOJI:
            desc = EMOJI[token.text]
            # Here we're converting shortcodes, e.g. ":man_getting_haircut:"
            return desc.replace("_", " ").replace(":", "")
        return None


def hashtag_getter(doc: Doc) -> List[str]:
    """Extract hashtags from text.

    Args:
        doc (Doc): A SpaCy document

    Returns:
        List[str]: A list of hashtags

    Example:
        >>> from spacy.tokens import Doc
        >>> Doc.set_extension("hashtag", getter=dacy.utilities.twitter.hashtags)
        >>> doc = nlp("Fuck hvor fedt! #yolo #life")
        >>> doc._.hashtag  # extrac the hashtags from your document
        ["#yolo", "#life"]
    """

    def find_hashtags(
        text,
        valid_tags={"#", "＃"},
        valid_chars={"_", "-"},
        invalid_tag_suffix={b"\xe2\x83\xa3", b"\xef\xb8\x8f"},
    ):
        def is_letter(t):
            if (
                t.isalnum()
                or t in valid_chars
                or str.encode(t).startswith(b"\xcc")
                or str.encode(t).startswith(b"\xe0")
            ):
                return True
            return False

        start = None
        for i, char in enumerate(text):
            if (
                char in valid_tags
                and not (
                    i + 1 != len(text) and str.encode(text[i + 1]) in invalid_tag_suffix
                )
                and (i == 0 or not (is_letter(text[i - 1]) or text[i - 1] == "&"))
            ):
                start = i
                continue
            if start is not None and not is_letter(char):
                if char in valid_tags:
                    start = None
                    continue
                print(start, i)
                if not text[start + 1 : i].isnumeric():
                    yield "#" + text[start + 1 : i]
                start = None
        if start is not None and not text[start + 1 : i + 1].isnumeric():
            print(start, i)
            yield "#" + text[start + 1 : i + 1]

    return list(find_hashtags(doc.text))


# def url_getter(doc: Doc) -> List[str]:
#     """
#     Extract hashtags from text

#     Args:
#         doc (Doc): A SpaCy document

#     Returns:
#         List[str]: A list of ulrs

#     Example:
#         >>> from spacy.tokens import Doc
#         >>> Doc.set_extension("hashtag", getter=dacy.utilities.twitter.hashtags)
#         >>> doc = nlp("Fuck hvor fedt! www.cool.dk")
#         >>> doc._.hashtag  # extrac the hashtags from your document
#         ["www.cool.dk"]
#     """
#     O = len(doc)
#     P = len(list(doc.sents))
#     L = len([t for t in doc if len(t) > 6])

#     LIX = O / P + L * 100 / O


#     return LIX


# def mentions_getter(doc: Doc) -> List[str]:
#     """
#     Extract hashtags from text

#     Args:
#         doc (Doc): A SpaCy document

#     Returns:
#         List[str]: A list of mentions

#     Example:
#         >>> from spacy.tokens import Doc
#         >>> Doc.set_extension("hashtag", getter=dacy.utilities.twitter.hashtags)
#         >>> doc = nlp("Fuck hvor fedt @arne!")
#         >>> doc._.hashtag  # extrac the hashtags from your document
#         ["@arne"]
#     """
#     O = len(doc)
#     P = len(list(doc.sents))
#     L = len([t for t in doc if len(t) > 6])

#     LIX = O / P + L * 100 / O


#     return LIX


def find_hashtags(
    text,
    valid_tags={"#", "＃"},
    valid_chars={"_", "-"},
    invalid_tag_suffix={b"\xe2\x83\xa3", b"\xef\xb8\x8f"},
):
    def is_letter(t):
        if (
            t.isalnum()
            or t in valid_chars
            or str.encode(t).startswith(b"\xcc")
            or str.encode(t).startswith(b"\xe0")
        ):
            return True
        return False

    start = None
    for i, char in enumerate(text):
        if (
            char in valid_tags
            and not (
                i + 1 != len(text) and str.encode(text[i + 1]) in invalid_tag_suffix
            )
            and (i == 0 or not (is_letter(text[i - 1]) or text[i - 1] == "&"))
        ):
            start = i
            continue
        if start is not None and not is_letter(char):
            if char in valid_tags:
                start = None
                continue
            print(start, i)
            if not text[start + 1 : i].isnumeric():
                yield "#" + text[start + 1 : i]
            start = None
    if start is not None and not text[start + 1 : i + 1].isnumeric():
        print(start, i)
        yield "#" + text[start + 1 : i + 1]


for s in [
    "Fuck hvor fedt! #yolo #life",
    "#taga",
    "invalid hashtags #double#tag",
    "invalid hashtags #double@tag",
    "also invalid #1",
]:
    result = list(find_hashtags(s))
    print(s, "\t", result)

# from spacy.lang.da import Danish
# nlp = Danish()
# doc = nlp("Fuck hvor fedt! #yolo #life")
# doc = nlp("invalid hashtags #double#tag")
# doc = nlp("invalid hashtags #double@trouble")
# city_getter = lambda doc: any(city in doc.text for city in ("New York", "Paris", "Berlin"))
# Doc.set_extension("has_city", getter=city_getter)


hash_tag_examples = """"
    - description: "Autolink trailing hashtag"
      text: "text #hashtag"
      expected: "text <a href=\"https://twitter.com/search?q=%23hashtag\" title=\"#hashtag\" class=\"tweet-url hashtag\">#hashtag</a>"

    - description: "Autolink alphanumeric hashtag (letter-number-letter)"
      text: "text #hash0tag"
      expected: "text <a href=\"https://twitter.com/search?q=%23hash0tag\" title=\"#hash0tag\" class=\"tweet-url hashtag\">#hash0tag</a>"

    - description: "Autolink alphanumeric hashtag (number-letter)"
      text: "text #1tag"
      expected: "text <a href=\"https://twitter.com/search?q=%231tag\" title=\"#1tag\" class=\"tweet-url hashtag\">#1tag</a>"

    - description: "Autolink hashtag with underscore"
      text: "text #hash_tag"
      expected: "text <a href=\"https://twitter.com/search?q=%23hash_tag\" title=\"#hash_tag\" class=\"tweet-url hashtag\">#hash_tag</a>"

    - description: "DO NOT Autolink all-numeric hashtags"
      text: "text #1234"
      expected: "text #1234"

    - description: "DO NOT Autolink hashtag preceded by a letter"
      text: "text#hashtag"
      expected: "text#hashtag"

    - description: "DO NOT Autolink hashtag that begins with \ufe0f (Emoji style hash sign)"
      text: "#️hashtag"
      expected: "#️hashtag"

    - description: "DO NOT Autolink hashtag that begins with \ufe0f (Keycap style hash sign)"
      text: "#⃣hashtag"
      expected: "#⃣hashtag"

    - description: "Autolink multiple hashtags"
      text: "text #hashtag1 #hashtag2"
      expected: "text <a href=\"https://twitter.com/search?q=%23hashtag1\" title=\"#hashtag1\" class=\"tweet-url hashtag\">#hashtag1</a> <a href=\"https://twitter.com/search?q=%23hashtag2\" title=\"#hashtag2\" class=\"tweet-url hashtag\">#hashtag2</a>"

    - description: "Autolink hashtag preceded by a period"
      text: "text.#hashtag"
      expected: "text.<a href=\"https://twitter.com/search?q=%23hashtag\" title=\"#hashtag\" class=\"tweet-url hashtag\">#hashtag</a>"

    - description: "DO NOT Autolink hashtag preceded by &"
      text: "&#nbsp;"
      expected: "&#nbsp;"

    - description: "Autolink hashtag followed by ! (! not included)"
      text: "text #hashtag!"
      expected: "text <a href=\"https://twitter.com/search?q=%23hashtag\" title=\"#hashtag\" class=\"tweet-url hashtag\">#hashtag</a>!"

    - description: "Autolink two hashtags separated by a slash"
      text: "text #dodge/#answer"
      expected: "text <a href=\"https://twitter.com/search?q=%23dodge\" title=\"#dodge\" class=\"tweet-url hashtag\">#dodge</a>/<a href=\"https://twitter.com/search?q=%23answer\" title=\"#answer\" class=\"tweet-url hashtag\">#answer</a>"

    - description: "Autolink hashtag before a slash"
      text: "text #dodge/answer"
      expected: "text <a href=\"https://twitter.com/search?q=%23dodge\" title=\"#dodge\" class=\"tweet-url hashtag\">#dodge</a>/answer"

    - description: "Autolink hashtag after a slash"
      text: "text dodge/#answer"
      expected: "text dodge/<a href=\"https://twitter.com/search?q=%23answer\" title=\"#answer\" class=\"tweet-url hashtag\">#answer</a>"

    - description: "Autolink hashtag followed by Japanese"
      text: "text #hashtagの"
      expected: "text <a href=\"https://twitter.com/search?q=%23hashtagの\" title=\"#hashtagの\" class=\"tweet-url hashtag\">#hashtagの</a>"

    - description: "Autolink hashtag preceded by full-width space (U+3000)"
      text: "text　#hashtag"
      expected: "text　<a href=\"https://twitter.com/search?q=%23hashtag\" title=\"#hashtag\" class=\"tweet-url hashtag\">#hashtag</a>"

    - description: "Autolink hashtag followed by full-width space (U+3000)"
      text: "#hashtag　text"
      expected: "<a href=\"https://twitter.com/search?q=%23hashtag\" title=\"#hashtag\" class=\"tweet-url hashtag\">#hashtag</a>　text"

    - description: "Autolink hashtag with full-width hash (U+FF03)"
      text: "＃hashtag"
      expected: "<a href=\"https://twitter.com/search?q=%23hashtag\" title=\"#hashtag\" class=\"tweet-url hashtag\">＃hashtag</a>"

    - description: "Autolink hashtag with accented character at the start"
      text: "#éhashtag"
      expected: "<a href=\"https://twitter.com/search?q=%23éhashtag\" title=\"#éhashtag\" class=\"tweet-url hashtag\">#éhashtag</a>"

    - description: "Autolink hashtag with accented character at the end"
      text: "#hashtagé"
      expected: "<a href=\"https://twitter.com/search?q=%23hashtagé\" title=\"#hashtagé\" class=\"tweet-url hashtag\">#hashtagé</a>"

    - description: "Autolink hashtag with accented character in the middle"
      text: "#hashétag"
      expected: "<a href=\"https://twitter.com/search?q=%23hashétag\" title=\"#hashétag\" class=\"tweet-url hashtag\">#hashétag</a>"

    - description: "Autolink hashtags in Korean"
      text: "What is #트위터 anyway?"
      expected: "What is <a href=\"https://twitter.com/search?q=%23트위터\" title=\"#트위터\" class=\"tweet-url hashtag\">#트위터</a> anyway?"

    - description: "Autolink hashtags in Russian"
      text: "What is #ашок anyway?"
      expected: "What is <a href=\"https://twitter.com/search?q=%23ашок\" title=\"#ашок\" class=\"tweet-url hashtag\">#ашок</a> anyway?"

    - description: "Autolink a katakana hashtag preceded by a space and followed by a space"
      text: "カタカナ #カタカナ カタカナ"
      expected: "カタカナ <a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a> カタカナ"

    - description: "Autolink a katakana hashtag preceded by a space and followed by a bracket"
      text: "カタカナ #カタカナ」カタカナ"
      expected: "カタカナ <a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a>」カタカナ"

    - description: "Autolink a katakana hashtag preceded by a space and followed by a edge"
      text: "カタカナ #カタカナ"
      expected: "カタカナ <a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a>"

    - description: "Autolink a katakana hashtag preceded by a bracket and followed by a space"
      text: "カタカナ「#カタカナ カタカナ"
      expected: "カタカナ「<a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a> カタカナ"

    - description: "Autolink a katakana hashtag preceded by a bracket and followed by a bracket"
      text: "カタカナ「#カタカナ」カタカナ"
      expected: "カタカナ「<a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a>」カタカナ"

    - description: "Autolink a katakana hashtag preceded by a bracket and followed by a edge"
      text: "カタカナ「#カタカナ"
      expected: "カタカナ「<a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a>"

    - description: "Autolink a katakana hashtag preceded by a edge and followed by a space"
      text: "#カタカナ カタカナ"
      expected: "<a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a> カタカナ"

    - description: "Autolink a katakana hashtag preceded by a edge and followed by a bracket"
      text: "#カタカナ」カタカナ"
      expected: "<a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a>」カタカナ"

    - description: "Autolink a katakana hashtag preceded by a edge and followed by a edge"
      text: "#カタカナ"
      expected: "<a href=\"https://twitter.com/search?q=%23カタカナ\" title=\"#カタカナ\" class=\"tweet-url hashtag\">#カタカナ</a>"

    - description: "Autolink a katakana hashtag with a voiced sounds mark followed by a space"
      text: "#ﾊｯｼｭﾀｸﾞ　テスト"
      expected: "<a href=\"https://twitter.com/search?q=%23ﾊｯｼｭﾀｸﾞ\" title=\"#ﾊｯｼｭﾀｸﾞ\" class=\"tweet-url hashtag\">#ﾊｯｼｭﾀｸﾞ</a>　テスト"

    - description: "Autolink a katakana hashtag with a voiced sounds mark followed by numbers"
      text: "#ﾊｯｼｭﾀｸﾞ123"
      expected: "<a href=\"https://twitter.com/search?q=%23ﾊｯｼｭﾀｸﾞ123\" title=\"#ﾊｯｼｭﾀｸﾞ123\" class=\"tweet-url hashtag\">#ﾊｯｼｭﾀｸﾞ123</a>"

    - description: "Autolink a katakana hashtag with another voiced sounds mark"
      text: "#ﾊﾟﾋﾟﾌﾟﾍﾟﾎﾟ"
      expected: "<a href=\"https://twitter.com/search?q=%23ﾊﾟﾋﾟﾌﾟﾍﾟﾎﾟ\" title=\"#ﾊﾟﾋﾟﾌﾟﾍﾟﾎﾟ\" class=\"tweet-url hashtag\">#ﾊﾟﾋﾟﾌﾟﾍﾟﾎﾟ</a>"

    - description: "Autolink a kanji hashtag preceded by a space and followed by a space"
      text: "漢字 #漢字 漢字"
      expected: "漢字 <a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a> 漢字"

    - description: "Autolink a kanji hashtag preceded by a space and followed by a bracket"
      text: "漢字 #漢字」漢字"
      expected: "漢字 <a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a>」漢字"

    - description: "Autolink a kanji hashtag preceded by a space and followed by a edge"
      text: "漢字 #漢字"
      expected: "漢字 <a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a>"

    - description: "Autolink a kanji hashtag preceded by a bracket and followed by a space"
      text: "漢字「#漢字 漢字"
      expected: "漢字「<a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a> 漢字"

    - description: "Autolink a kanji hashtag preceded by a bracket and followed by a bracket"
      text: "漢字「#漢字」漢字"
      expected: "漢字「<a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a>」漢字"

    - description: "Autolink a kanji hashtag preceded by a bracket and followed by a edge"
      text: "漢字「#漢字"
      expected: "漢字「<a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a>"

    - description: "Autolink a kanji hashtag preceded by a edge and followed by a space"
      text: "#漢字 漢字"
      expected: "<a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a> 漢字"

    - description: "Autolink a kanji hashtag preceded by a edge and followed by a bracket"
      text: "#漢字」漢字"
      expected: "<a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a>」漢字"

    - description: "Autolink a kanji hashtag preceded by a edge and followed by a edge"
      text: "#漢字"
      expected: "<a href=\"https://twitter.com/search?q=%23漢字\" title=\"#漢字\" class=\"tweet-url hashtag\">#漢字</a>"

    - description: "Autolink a kanji hashtag preceded by an ideographic comma, followed by an ideographic period"
      text: "これは、＃大丈夫。"
      expected: "これは、<a href=\"https://twitter.com/search?q=%23大丈夫\" title=\"#大丈夫\" class=\"tweet-url hashtag\">＃大丈夫</a>。"

    - description: "Autolink a hiragana hashtag preceded by a space and followed by a space"
      text: "ひらがな #ひらがな ひらがな"
      expected: "ひらがな <a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a> ひらがな"

    - description: "Autolink a hiragana hashtag preceded by a space and followed by a bracket"
      text: "ひらがな #ひらがな」ひらがな"
      expected: "ひらがな <a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a>」ひらがな"

    - description: "Autolink a hiragana hashtag preceded by a space and followed by a edge"
      text: "ひらがな #ひらがな"
      expected: "ひらがな <a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a>"

    - description: "Autolink a hiragana hashtag preceded by a bracket and followed by a space"
      text: "ひらがな「#ひらがな ひらがな"
      expected: "ひらがな「<a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a> ひらがな"

    - description: "Autolink a hiragana hashtag preceded by a bracket and followed by a bracket"
      text: "ひらがな「#ひらがな」ひらがな"
      expected: "ひらがな「<a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a>」ひらがな"

    - description: "Autolink a hiragana hashtag preceded by a bracket and followed by a edge"
      text: "ひらがな「#ひらがな"
      expected: "ひらがな「<a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a>"

    - description: "Autolink a hiragana hashtag preceded by a edge and followed by a space"
      text: "#ひらがな ひらがな"
      expected: "<a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a> ひらがな"

    - description: "Autolink a hiragana hashtag preceded by a edge and followed by a bracket"
      text: "#ひらがな」ひらがな"
      expected: "<a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a>」ひらがな"

    - description: "Autolink a hiragana hashtag preceded by a edge and followed by a edge"
      text: "#ひらがな"
      expected: "<a href=\"https://twitter.com/search?q=%23ひらがな\" title=\"#ひらがな\" class=\"tweet-url hashtag\">#ひらがな</a>"

    - description: "Autolink a Kanji/Katakana mix hashtag"
      text: "日本語ハッシュタグ #日本語ハッシュタグ"
      expected: "日本語ハッシュタグ <a href=\"https://twitter.com/search?q=%23日本語ハッシュタグ\" title=\"#日本語ハッシュタグ\" class=\"tweet-url hashtag\">#日本語ハッシュタグ</a>"

    - description: "DO NOT autolink a hashtag without a preceding space"
      text: "日本語ハッシュタグ#日本語ハッシュタグ"
      expected: "日本語ハッシュタグ#日本語ハッシュタグ"

    - description: "DO NOT include a punctuation in a hashtag"
      text: "#日本語ハッシュタグ。"
      expected: "<a href=\"https://twitter.com/search?q=%23日本語ハッシュタグ\" title=\"#日本語ハッシュタグ\" class=\"tweet-url hashtag\">#日本語ハッシュタグ</a>。"

    - description: "Autolink a hashtag after a punctuation"
      text: "日本語ハッシュタグ。#日本語ハッシュタグ"
      expected: "日本語ハッシュタグ。<a href=\"https://twitter.com/search?q=%23日本語ハッシュタグ\" title=\"#日本語ハッシュタグ\" class=\"tweet-url hashtag\">#日本語ハッシュタグ</a>"

    - description: "Autolink a hashtag with chouon"
      text: "長音ハッシュタグ。#サッカー"
      expected: "長音ハッシュタグ。<a href=\"https://twitter.com/search?q=%23サッカー\" title=\"#サッカー\" class=\"tweet-url hashtag\">#サッカー</a>"

    - description: "Autolink a hashtag with half-width chouon"
      text: "長音ハッシュタグ。#ｻｯｶｰ"
      expected: "長音ハッシュタグ。<a href=\"https://twitter.com/search?q=%23ｻｯｶｰ\" title=\"#ｻｯｶｰ\" class=\"tweet-url hashtag\">#ｻｯｶｰ</a>"

    - description: "Autolink a hashtag with half-width # after full-width ！"
      text: "できましたよー！#日本語ハッシュタグ。"
      expected: "できましたよー！<a href=\"https://twitter.com/search?q=%23日本語ハッシュタグ\" title=\"#日本語ハッシュタグ\" class=\"tweet-url hashtag\">#日本語ハッシュタグ</a>。"

    - description: "Autolink a hashtag with full-width ＃ after full-width ！"
      text: "できましたよー！＃日本語ハッシュタグ。"
      expected: "できましたよー！<a href=\"https://twitter.com/search?q=%23日本語ハッシュタグ\" title=\"#日本語ハッシュタグ\" class=\"tweet-url hashtag\">＃日本語ハッシュタグ</a>。"

    - description: "Autolink a hashtag containing ideographic iteration mark"
      text: "#云々"
      expected: "<a href=\"https://twitter.com/search?q=%23云々\" title=\"#云々\" class=\"tweet-url hashtag\">#云々</a>"

    - description: "Autolink multiple hashtags in multiple languages"
      text: "Hashtags in #中文, #日本語, #한국말, and #русский! Try it out!"
      expected: "Hashtags in <a href=\"https://twitter.com/search?q=%23中文\" title=\"#中文\" class=\"tweet-url hashtag\">#中文</a>, <a href=\"https://twitter.com/search?q=%23日本語\" title=\"#日本語\" class=\"tweet-url hashtag\">#日本語</a>, <a href=\"https://twitter.com/search?q=%23한국말\" title=\"#한국말\" class=\"tweet-url hashtag\">#한국말</a>, and <a href=\"https://twitter.com/search?q=%23русский\" title=\"#русский\" class=\"tweet-url hashtag\">#русский</a>! Try it out!"

    - description: "Autolink should allow for ş (U+015F) in a hashtag"
      text: "Here’s a test tweet for you: #Ateş #qrşt #ştu #ş"
      expected: "Here’s a test tweet for you: <a href=\"https://twitter.com/search?q=%23Ateş\" title=\"#Ateş\" class=\"tweet-url hashtag\">#Ateş</a> <a href=\"https://twitter.com/search?q=%23qrşt\" title=\"#qrşt\" class=\"tweet-url hashtag\">#qrşt</a> <a href=\"https://twitter.com/search?q=%23ştu\" title=\"#ştu\" class=\"tweet-url hashtag\">#ştu</a> <a href=\"https://twitter.com/search?q=%23ş\" title=\"#ş\" class=\"tweet-url hashtag\">#ş</a>"

    - description: "Autolink a hashtag with Latin extended character"
      text: "#mûǁae"
      expected: "<a href=\"https://twitter.com/search?q=%23mûǁae\" title=\"#mûǁae\" class=\"tweet-url hashtag\">#mûǁae</a>"

# Please be careful with changes to this test case - what looks like "á" is really a + U+0301, and many editors will silently convert this to U+00E1.
    - description: "Autolink hashtags with combining diacritics"
      text: "#táim #hag̃ua"
      expected: "<a href=\"https://twitter.com/search?q=%23táim\" title=\"#táim\" class=\"tweet-url hashtag\">#táim</a> <a href=\"https://twitter.com/search?q=%23hag̃ua\" title=\"#hag̃ua\" class=\"tweet-url hashtag\">#hag̃ua</a>"

    - description: "Autolink Arabic hashtag"
      text: "Arabic hashtag: #فارسی #لس_آنجلس"
      expected: "Arabic hashtag: <a href=\"https://twitter.com/search?q=%23فارسی\" title=\"#فارسی\" class=\"tweet-url hashtag rtl\">#فارسی</a> <a href=\"https://twitter.com/search?q=%23لس_آنجلس\" title=\"#لس_آنجلس\" class=\"tweet-url hashtag rtl\">#لس_آنجلس</a>"

    - description: "Autolink Thai hashtag"
      text: "Thai hashtag: #รายละเอียด"
      expected: "Thai hashtag: <a href=\"https://twitter.com/search?q=%23รายละเอียด\" title=\"#รายละเอียด\" class=\"tweet-url hashtag\">#รายละเอียด</a>"
"""

import re

lines = hash_tag_examples.split("\n")
examples = []
for i, line in enumerate(lines):
    if line.startswith("    - description"):
        rex = re.compile(r'<a.*? title="(.*?)" .*?<\/a>')
        match = rex.findall(lines[i + 2][17:-1])
        examples.append(
            {
                "desc": lines[i],
                "text": lines[i + 1][13:-1],
                "expected": lines[i + 2][17:-1],
                "hashtags": match,
            },
        )


for i, d in enumerate(examples):
    print("i", i)
    hashtags = list(find_hashtags(d["text"]))
    for h in hashtags:
        if h not in d["hashtags"]:
            raise ValueError
    for h in d["hashtags"]:
        if h not in hashtags:
            raise Exception

d["desc"]
d["text"]
d["expected"]
hashtags
str.encode(d["text"][0]) in {str.encode("#")}
str.encode("#️")
str.encode("#hashtags")

"#️⃣"

list(find_hashtags("#️⃣yolo"))

for i in d["text"]:
    print(i, i.isalnum())
str.encode(d["text"][22])

str.encode("áim"[1])
"â"[0]


hashtag_tests = [
    ("text #hashtag", ["#hashtag"]),
    ("text #hash0tag", ["#hash0tag"]),
    ("text #1tag", ["#1tag"]),
    ("text #hash_tag", ["#hash_tag"]),
    ("text #1234", []),
    ("text#hashtag", []),
    ("#️hashtag", []),
    ("#⃣hashtag", []),
    ("text #hashtag1 #hashtag2", ["#hashtag1", "#hashtag2"]),
    ("text.#hashtag", ["#hashtag"]),
    ("&#nbsp;", []),
    ("text #hashtag!", ["#hashtag"]),
    ("text #dodge/#answer", ["#dodge", "#answer"]),
    ("text #dodge/answer", ["#dodge"]),
    ("text dodge/#answer", ["#answer"]),
    ("text #hashtagの", ["#hashtagの"]),
    ("text\u3000#hashtag", ["#hashtag"]),
    ("#hashtag\u3000text", ["#hashtag"]),
    ("＃hashtag", ["#hashtag"]),
    ("#éhashtag", ["#éhashtag"]),
    ("#hashtagé", ["#hashtagé"]),
    ("#hashétag", ["#hashétag"]),
    ("What is #트위터 anyway?", ["#트위터"]),
    ("What is #ашок anyway?", ["#ашок"]),
    ("カタカナ #カタカナ カタカナ", ["#カタカナ"]),
    ("カタカナ #カタカナ」カタカナ", ["#カタカナ"]),
    ("カタカナ #カタカナ", ["#カタカナ"]),
    ("カタカナ「#カタカナ カタカナ", ["#カタカナ"]),
    ("カタカナ「#カタカナ」カタカナ", ["#カタカナ"]),
    ("カタカナ「#カタカナ", ["#カタカナ"]),
    ("#カタカナ カタカナ", ["#カタカナ"]),
    ("#カタカナ」カタカナ", ["#カタカナ"]),
    ("#カタカナ", ["#カタカナ"]),
    ("#ﾊｯｼｭﾀｸﾞ\u3000テスト", ["#ﾊｯｼｭﾀｸﾞ"]),
    ("#ﾊｯｼｭﾀｸﾞ123", ["#ﾊｯｼｭﾀｸﾞ123"]),
    ("#ﾊﾟﾋﾟﾌﾟﾍﾟﾎﾟ", ["#ﾊﾟﾋﾟﾌﾟﾍﾟﾎﾟ"]),
    ("漢字 #漢字 漢字", ["#漢字"]),
    ("漢字 #漢字」漢字", ["#漢字"]),
    ("漢字 #漢字", ["#漢字"]),
    ("漢字「#漢字 漢字", ["#漢字"]),
    ("漢字「#漢字」漢字", ["#漢字"]),
    ("漢字「#漢字", ["#漢字"]),
    ("#漢字 漢字", ["#漢字"]),
    ("#漢字」漢字", ["#漢字"]),
    ("#漢字", ["#漢字"]),
    ("これは、＃大丈夫。", ["#大丈夫"]),
    ("ひらがな #ひらがな ひらがな", ["#ひらがな"]),
    ("ひらがな #ひらがな」ひらがな", ["#ひらがな"]),
    ("ひらがな #ひらがな", ["#ひらがな"]),
    ("ひらがな「#ひらがな ひらがな", ["#ひらがな"]),
    ("ひらがな「#ひらがな」ひらがな", ["#ひらがな"]),
    ("ひらがな「#ひらがな", ["#ひらがな"]),
    ("#ひらがな ひらがな", ["#ひらがな"]),
    ("#ひらがな」ひらがな", ["#ひらがな"]),
    ("#ひらがな", ["#ひらがな"]),
    ("日本語ハッシュタグ #日本語ハッシュタグ", ["#日本語ハッシュタグ"]),
    ("日本語ハッシュタグ#日本語ハッシュタグ", []),
    ("#日本語ハッシュタグ。", ["#日本語ハッシュタグ"]),
    ("日本語ハッシュタグ。#日本語ハッシュタグ", ["#日本語ハッシュタグ"]),
    ("長音ハッシュタグ。#サッカー", ["#サッカー"]),
    ("長音ハッシュタグ。#ｻｯｶｰ", ["#ｻｯｶｰ"]),
    ("できましたよー！#日本語ハッシュタグ。", ["#日本語ハッシュタグ"]),
    ("できましたよー！＃日本語ハッシュタグ。", ["#日本語ハッシュタグ"]),
    ("#云々", ["#云々"]),
    (
        "Hashtags in #中文, #日本語, #한국말, and #русский! Try it out!",
        ["#中文", "#日本語", "#한국말", "#русский"],
    ),
    (
        "Here’s a test tweet for you: #Ateş #qrşt #ştu #ş",
        ["#Ateş", "#qrşt", "#ştu", "#ş"],
    ),
    ("#mûǁae", ["#mûǁae"]),
    ("#táim #hag̃ua", ["#táim", "#hag̃ua"]),
    ("Arabic hashtag: #فارسی #لس_آنجلس", ["#فارسی", "#لس_آنجلس"]),
    ("Thai hashtag: #รายละเอียด", ["#รายละเอียด"]),
]



"""Contains functions for testing the performance of models on varying input
length."""
from typing import Callable, List, Union

import pandas as pd
from wasabi import msg

from ..datasets import dane
from .score import score


def n_sents_score(
    n_sents: Union[int, List[int]],
    apply_fn: Callable,
    dataset: str = "dane",
    split: str = "test",
    score_fn: List[Union[str, Callable]] = ["token", "pos", "ents", "dep"],
    verbose: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """scores the performance of a given model on examples of a given number of
    sentences.

    Args:
        n_sents (Union[int, List[int]]): Number of sentences which the performance
            should be applied to.
        apply_fn (Callable):  A wrapper function for the model you wish to score. The
            model should take in a spacy Example and output a tagged version of it.
        dataset (str, optional): Which dataset should this be applied to. Possible
            options include "dane". Defaults to "dane".
        split (str, optional): Which splits of the dataset should be used. Possible
            options include "train", "dev", "test", "all". Defaults to "test".
        score_fn (List[Union[str, Callable]], optional): A scoring function which takes
            in a list of examples and return a dictionary of the form {"score_name":
            score}. Four potiential strings are valid. "ents" for measuring the
            performance of entity spans. "pos" for measuring the performance of
            pos-tags. "token" for measuring the performance of tokenization. "nlp" for
            measuring the performance of all components in the specified nlp pipeline.
            Defaults to ["token", "pos", "ents"].
        verbose (bool, optional): Toggles the verbosity of the function. Defualts to
            True
        kwargs (dict): arguments to be passed to dataset or the score function.

    Returns:
        pandas.DataFrame: returns a pandas dataframe containing the performance metrics.
    """

    dataset_fn = {"dane": dane}
    if isinstance(n_sents, int):
        n_sents = [n_sents]

    for i, n in enumerate(n_sents):
        if verbose is True:
            msg.info(f"Calculating score using {n} sentences")
        corpus = dataset_fn[dataset](splits=split, n_sents=n, **kwargs)
        scores_ = score(corpus, apply_fn=apply_fn, score_fn=score_fn, **kwargs)
        scores = pd.concat([scores, scores_]) if i != 0 else scores_  # noqa
    return scores



import pandas as pd
from spacy.lang.da import Danish
from spacy.training.augment import create_lower_casing_augmenter

from dacy.datasets import dane
from dacy.score import n_sents_score, score


def test_score():
    nlp = Danish()

    def apply_model(examples):
        e = []
        for example in examples:
            example.predicted = nlp(example.predicted.text)
            e.append(example)
        return e

    test = dane(splits=["test"])
    test.limit = 1

    scores = score(
        corpus=test,
        augmenter=create_lower_casing_augmenter(0.5),
        apply_fn=apply_model,
        k=3,
        score_fn=["ents", "pos", "token"],
    )
    assert isinstance(scores, pd.DataFrame)

    # test with nlp as input
    scores_ = score(
        corpus=test,
        augmenter=create_lower_casing_augmenter(0.5),
        apply_fn=nlp,
        k=3,
        score_fn=["ents", "pos", "token"],
    )
    for s in scores_:
        assert s in scores.columns


def test_n_sents_score():
    nlp = Danish()

    def apply_model(examples):
        e = []
        for example in examples:
            example.predicted = nlp(example.predicted.text)
            e.append(example)
        return e

    scores = n_sents_score(
        n_sents=1,
        apply_fn=apply_model,
    )
    assert isinstance(scores, pd.DataFrame)



"""Functions for downloading DaCy models."""
import os
import shutil
from pathlib import Path
from typing import Optional

from tqdm import tqdm
from wasabi import msg

DEFAULT_CACHE_DIR = os.getenv("DACY_CACHE_DIR", os.path.join(str(Path.home()), ".dacy"))

models_url = {
    "da_dacy_small_tft-0.0.0": "https://sciencedata.dk//shared/d845d4fef9ea165ee7bd6dd954b95de2?download",  # noqa: E501
    "da_dacy_medium_tft-0.0.0": "https://sciencedata.dk//shared/c205edf59195583122d7213a3c26c077?download",  # noqa: E501
    "da_dacy_large_tft-0.0.0": "https://sciencedata.dk//shared/0da7cb975b245d9e6574458c7c89dfd9?download",  # noqa: E501
    "da_dacy_small_trf-0.1.0": "https://huggingface.co/chcaa/da_dacy_small_trf/resolve/main/da_dacy_small_trf-any-py3-none-any.whl",  # noqa: E501
    "da_dacy_medium_trf-0.1.0": "https://huggingface.co/chcaa/da_dacy_medium_trf/resolve/main/da_dacy_medium_trf-any-py3-none-any.whl",  # noqa: E501
    "da_dacy_large_trf-0.1.0": "https://huggingface.co/chcaa/da_dacy_large_trf/resolve/main/da_dacy_large_trf-any-py3-none-any.whl",  # noqa: E501
}


def models() -> list:
    """Returns a list of valid DaCy models.

    Returns:
        list: list of valid DaCy models
    """
    return list(models_url.keys())


class DownloadProgressBar(tqdm):
    def update_to(self, b: int = 1, bsize: int = 1, tsize=None) -> None:
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url: str, output_path: str) -> None:
    import urllib.request

    with DownloadProgressBar(
        unit="B",
        unit_scale=True,
        miniters=1,
        desc=url.split("/")[-1],
    ) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def install(package):
    import subprocess
    import sys

    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", package, "--no-deps"],
    )


def download_model(
    model: str,
    save_path: Optional[str] = None,
    force: bool = False,
    verbose: bool = True,
) -> str:
    """Downloads and install a specified DaCy pipeline.

    Args:
        model (str): string indicating DaCy model, use dacy.models() to get a list of
            models
        save_path (str, optional): The path you want to save your model to. Is only used
            for DaCy models of v0.0.0 as later models are installed as modules to allow
            for better versioning. Defaults to None denoting the default cache
            directory. This can be found using using dacy.where_is_my_dacy().
        force (bool, optional): Should it download the model regardless of it already
            being present? Defaults to False.
        verbose (bool): Toggles the verbosity of the function. Defaults to True.

    Returns:
        a string of the model location

    Example:
        >>> download_model(model="da_dacy_medium_tft-0.0.0")
    """
    if model in {"small", "medium", "large"}:
        model = f"da_dacy_{model}_trf-0.1.0"

    if model not in models_url:
        raise ValueError(
            "The model is not available in DaCy. Please use dacy.models() to see a"
            + " list of all models",
        )

    if int(model.split("-")[-1].split(".")[1]) < 1:  # model v. 0.0.0
        if save_path is None:
            save_path = DEFAULT_CACHE_DIR

        url = models_url[model]
        path = os.path.join(save_path, model)
        dl_path = os.path.join(save_path, "tmp.zip")
        if os.path.exists(path) and force is False:
            return path

        if verbose is True:
            msg.info(f"\nDownloading '{model}'")
        Path(save_path).mkdir(parents=True, exist_ok=True)
        download_url(url, dl_path)
        shutil.unpack_archive(dl_path, save_path)
        os.remove(dl_path)
        if verbose is True:
            msg.info(
                r"\Model successfully downloaded, you can now load it using dacy. "
                + f"load({model})",
            )
        return path
    else:
        from spacy.util import get_installed_models

        mdl = model.split("-")[0]
        if mdl not in get_installed_models():
            install(models_url[model])
        return mdl



from typing import Callable, List, Optional
from warnings import warn

from spacy.lang.da import Danish
from spacy.language import Language
from spacy.tokens import Doc
from spacy_transformers.data_classes import FullTransformerBatch
from spacy_wrap import (
    SequenceClassificationTransformer,
    make_sequence_classification_transformer,
)
from thinc.api import Config, Model

DEFAULT_CONFIG_STR = """
[hatespeech_detection]
max_batch_items = 4096
doc_extension_trf_data = "is_offensive_trf_data"
doc_extension_prediction = "is_offensive"
labels = ["not offensive", "offensive"]

[hatespeech_detection.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[hatespeech_detection.model]
@architectures = "spacy-wrap.SequenceClassificationTransformerModel.v1"
name = "alexandrainst/da-hatespeech-detection-base"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}

[hatespeech_detection.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96


[hatespeech_classification]
max_batch_items = 4096
doc_extension_trf_data = "hate_speech_type_trf_data"
doc_extension_prediction = "hate_speech_type"
labels = ["særlig opmærksomhed", "personangreb", "sprogbrug", "spam & indhold"]

[hatespeech_classification.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[hatespeech_classification.model]
@architectures = "spacy-wrap.SequenceClassificationTransformerModel.v1"
name = "alexandrainst/da-hatespeech-classification-base"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}

[hatespeech_classification.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96
"""


DEFAULT_CONFIG = Config().from_str(DEFAULT_CONFIG_STR)


Danish.factory(
    "dacy/hatespeech_detection",
    default_config=DEFAULT_CONFIG["hatespeech_detection"],
)(make_sequence_classification_transformer)


@Danish.factory(
    "dacy/hatespeech_classification",
    default_config=DEFAULT_CONFIG["hatespeech_classification"],
)
def make_offensive_transformer(
    nlp: Language,
    name: str,
    model: Model[List[Doc], FullTransformerBatch],
    set_extra_annotations: Callable[[List[Doc], FullTransformerBatch], None],
    max_batch_items: int,
    doc_extension_trf_data: str,
    doc_extension_prediction: str,
    labels: List[str],
) -> SequenceClassificationTransformer:
    if not Doc.has_extension("is_offensive"):
        warn(
            "The component assumes the 'is_offensive' extension is set."
            + " To set it you can run  nlp.add_pipe('dacy/hatespeech_detection')",
        )

    # TODO: Add a conditional forward such that the model isn't run is document is not
    # emotionally laden
    clf_mdl = SequenceClassificationTransformer(
        vocab=nlp.vocab,
        model=model,
        set_extra_annotations=set_extra_annotations,
        max_batch_items=max_batch_items,
        name=name,
        labels=labels,
        doc_extension_trf_data=doc_extension_trf_data,
        doc_extension_prediction=doc_extension_prediction,
        assign_to_cats=True,
    )

    # overwrite extension such that it return not offensive if the document is not
    # offensive
    if Doc.has_extension("is_offensive"):

        def label_getter(doc) -> Optional[str]:
            if doc._.is_offensive == "offensive":
                prob = getattr(doc._, f"{doc_extension_prediction}_prob")
                if prob["prob"] is not None:
                    return labels[int(prob["prob"].argmax())]
            return doc._.is_offensive

        Doc.set_extension(doc_extension_prediction, getter=label_getter, force=True)
    clf_mdl.model.initialize()

    return clf_mdl



from spacy.lang.da import Danish
from spacy_wrap.pipeline_component_tok_clf import make_token_classification_transformer
from thinc.api import Config

DEFAULT_CONFIG_STR = """
[token_classification_transformer]
max_batch_items = 4096
doc_extension_trf_data = "tok_clf_trf_data"
doc_extension_prediction = "tok_clf_predictions"
predictions_to = null
labels = null
aggregation_strategy = "first"

[token_classification_transformer.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[token_classification_transformer.model]
@architectures = "spacy-wrap.TokenClassificationTransformerModel.v1"
name="saattrupdan/nbailab-base-ner-scandi"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}


[token_classification_transformer.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96
"""


DEFAULT_CONFIG = Config().from_str(DEFAULT_CONFIG_STR)


Danish.factory(
    "dacy/ner",
    default_config=DEFAULT_CONFIG["token_classification_transformer"],
)(make_token_classification_transformer)



### DaNLP's BERT model requires transformers==3.5.1 (install with pip install transformers==3.5.1 --no-deps)

# to download the danlp and nerda you will have to set up a certificate:
import ssl

from danlp.models import load_bert_ner_model
from spacy.lang.da import Danish
from spacy.training import Example

from .apply_fn_utils import add_iob, apply_on_multiple_examples, no_misc_getter

ssl._create_default_https_context = ssl._create_unverified_context

bert_model = load_bert_ner_model()
nlp_da = Danish()


def __apply_bert_model(example: Example) -> Example:
    # uses spacy tokenization
    doc = nlp_da(example.reference.text)
    tokens, labels = bert_model.predict([t.text for t in doc])
    doc = add_iob(doc, labels)
    return Example(doc, example.reference)


apply_danlp_bert = apply_on_multiple_examples(__apply_bert_model)


if __name__ == "__main__":
    import os

    os.chdir("..")
    from dacy.datasets import dane

    test = dane(splits=["test"])
    nlp = Danish()
    examples = apply_danlp_bert(test(nlp))

    from spacy.scorer import Scorer

    tok_scores = Scorer.score_tokenization(examples)
    ent_scores = Scorer.score_spans(
        examples=examples,
        attr="ents",
        getter=no_misc_getter,
    )
    pos_scores = Scorer.score_token_attr(examples, "tag")

    from spacy import displacy

    displacy.render(examples[0].y, style="ent")
    displacy.render(examples[0].x, style="ent")



"""python -m spacy train config.cfg --code functions.py."""

import sys
from typing import Callable, Iterable, Iterator

import spacy

sys.path.append("../..")

from spacy.language import Language
from spacy.training import Example
from spacy.training.augment import create_lower_casing_augmenter

from dacy.augmenters import (
    create_char_swap_augmenter,
    create_keyboard_augmenter,
    create_pers_augmenter,
    create_æøå_augmenter,
)
from dacy.datasets.names import load_names


def combine_augmenters(
    augmenters: Iterable[Callable[[Language, Example], Iterator[Example]]],
) -> Callable[[Language, Example], Iterator[Example]]:
    """Combines a series og spaCy style augmenters.

    Args:
        augmenters (Iterable[Callable[[Language, Example], Iterator[Example]]]): An list of spaCy augmenters.

    Returns:
        Callable[[Language, Example], Iterator[Example]]: The combined augmenter
    """

    def apply_multiple_augmenters(nlp: Language, example: Example):
        examples = [example]
        for aug in augmenters:
            examples = (e for example in examples for e in aug(nlp, example))
        for e in examples:
            yield example

    return apply_multiple_augmenters


@spacy.registry.augmenters("dacy_augmenter.v1")
def dacy_augmenters():
    augmenters = [
        create_keyboard_augmenter(doc_level=1, char_level=0.02, distance=1.5),
        create_pers_augmenter(
            ent_dict=load_names(),
            patterns=[
                "fn,ln",
                "abbpunct,ln",
                "fn,ln,ln",
                "fn,ln,ln",
                "abb,ln",
                "ln,abbpunct",
            ],
            force_pattern_size=True,
            keep_name=False,
            prob=0.1,
        ),
        create_char_swap_augmenter(doc_level=1, char_level=0.02),
        create_æøå_augmenter(doc_level=1, char_level=0.1),
        create_lower_casing_augmenter(level=0.1),
    ]
    return combine_augmenters(augmenters)



### pip install daluke==0.0.5
from typing import Iterable, List

from daluke import AutoNERDaLUKE, predict_ner
from spacy.lang.da import Danish
from spacy.tokens import Doc
from spacy.training import Example

from .apply_fn_utils import add_iob, no_misc_getter

# This also downloads daluke model (first time)
daluke = AutoNERDaLUKE()

nlp_da = Danish()


def apply_daluke(
    examples: Iterable[Example],
    use_spacy: bool = True,
    batch_size: int = 16,
) -> List[Example]:
    docs_y, sentences = list(), list()
    for example in examples:
        # Tokenization using spacy or nltk
        if use_spacy:
            sentences.append([t.text for t in nlp_da(example.reference.text)])
        else:
            from nltk.tokenize import word_tokenize

            sentences.append(word_tokenize(example.reference.text))
        docs_y.append(example.reference)
    # NER using daluke
    # join `should` not give size issues, as this string is again crudely split in DaLUKE API
    predictions = predict_ner(
        [" ".join(sent) for sent in sentences],
        daluke,
        batch_size=batch_size,
    )
    out_examples = list()
    for doc_y, pred, words in zip(docs_y, predictions, sentences):
        doc = add_iob(Doc(nlp_da.vocab, words=words), iob=pred)
        out_examples.append(Example(doc, doc_y))
    return out_examples


if __name__ == "__main__":
    import os

    os.chdir("..")
    from dacy.datasets import dane

    test = dane(splits=["test"])
    nlp = Danish()
    examples = apply_daluke(test(nlp))

    from spacy.scorer import Scorer

    tok_scores = Scorer.score_tokenization(examples)
    ent_scores = Scorer.score_spans(
        examples=examples,
        attr="ents",
        getter=no_misc_getter,
    )
    pos_scores = Scorer.score_token_attr(examples, "tag")

    from spacy import displacy

    displacy.render(examples[0].y, style="ent")
    displacy.render(examples[0].x, style="ent")
    breakpoint()



import urllib

import dacy
from dacy.download import download_model, models_url
from dacy.load import load


def test_urls():
    for m in dacy.models():
        url = models_url[m]
        req = urllib.request.Request(url, method="HEAD")
        f = urllib.request.urlopen(req)
        assert f.status == 200

    for m in [
        "da_dacy_small_tft-0.0.0",
        "da_dacy_medium_tft-0.0.0",
        "da_dacy_large_tft-0.0.0",
    ]:
        print("\t Status:", f.status)
        size = int(f.headers["Content-Length"]) / 1e6
        assert size > 20
        print("\t File Size:", round(size), "mb")


def test_load():
    models = ["da_dacy_medium_trf-0.1.0", "da_dacy_medium_tft-0.0.0"]
    for m in models:
        nlp = load(m)
        nlp("Dette er en test tekst")


def test_models():
    print(dacy.models())


def test_where_is_my_dacy():
    print(dacy.where_is_my_dacy())


def test_download_model_error():
    download_model(model="da_dacy_medium_tft-0.0.0", force=True)
    # this just tests if it fails when needed.
    try:
        download_model(model="not a dacy model")
    except ValueError:
        pass



from typing import Callable, Iterable, List

from spacy.tokens import Doc, Span
from spacy.training import Example


def no_misc_getter(doc, attr):
    spans = getattr(doc, attr)
    for span in spans:
        if span.label_ == "MISC":
            continue
        yield span


def add_iob(doc: Doc, iob: List[str]) -> Doc:
    """Add iob tags to Doc.

    Args:
        doc (Doc): A SpaCy doc
        iob (List[str]): a list of tokens on the IOB format

    Returns:
        Doc: A doc with the spans to the new IOB
    """
    ent = []
    for i, label in enumerate(iob):
        # turn OOB labels into spans
        if label == "O":
            continue
        iob_, ent_type = label.split("-")
        if (i - 1 >= 0 and iob_ == "I" and iob[i - 1] == "O") or (
            i == 0 and iob_ == "I"
        ):
            iob_ = "B"
        if iob_ == "B":
            start = i
        if i + 1 >= len(iob) or iob[i + 1].split("-")[0] != "I":
            ent.append(Span(doc, start, i + 1, label=ent_type))
    doc.set_ents(ent)
    return doc


def apply_on_multiple_examples(func: Callable) -> Callable:
    def inner(examples: Iterable[Example], **kwargs) -> Iterable[Example]:
        return [func(e, **kwargs) for e in examples]

    return inner



# pip install flair==0.4.5

from typing import Iterable, List

from danlp.models import load_flair_ner_model, load_flair_pos_model
from flair.data import Sentence, Token
from spacy.lang.da import Danish
from spacy.tokens import Doc
from spacy.training import Example

tagger_ner = load_flair_ner_model()
tagger_pos = load_flair_pos_model()
nlp_da = Danish()


def apply_flair(examples: Iterable[Example], use_spacy=True) -> List[Example]:
    flair_sentences = []
    docs_y = []
    for example in examples:
        # tokenization
        if use_spacy:
            sent = Sentence()
            [sent.add_token(Token(t.text)) for t in nlp_da(example.reference.text)]
        else:
            sent = Sentence(example.reference.text)

        flair_sentences.append(sent)

        docs_y.append(example.reference)

    tagger_ner.predict(flair_sentences, verbose=True)
    tagger_pos.predict(flair_sentences, verbose=True)

    examples_ = []
    for doc_y, f_sent in zip(docs_y, flair_sentences):
        text, iob, upos, ws = zip(
            *[
                (
                    tok.text,
                    tok.tags["ner"].value,
                    tok.tags["upos"].value,
                    tok.whitespace_after,
                )
                for tok in f_sent
            ]
        )
        doc = Doc(nlp_da.vocab, words=text, spaces=ws, tags=upos, ents=iob)
        examples_.append(Example(doc, doc_y))
    return examples_


if __name__ == "__main__":
    import os

    os.chdir("..")
    from dacy.datasets import dane

    test = dane(splits=["test"])
    examples = apply_flair(test(nlp_da), use_spacy=False)
    examples_spacy = apply_flair(test(nlp_da), use_spacy=True)

    from spacy import displacy

    displacy.render(examples[5].y, style="ent")
    displacy.render(examples[5].x, style="ent")

    from spacy.scorer import Scorer

    from .apply_fn_utils import no_misc_getter

    tok_scores = Scorer.score_tokenization(examples)

    ent_scores = Scorer.score_spans(
        examples=examples,
        attr="ents",
        allow_overlap=True,
        getter=no_misc_getter,
    )
    ent_scores_spacy = Scorer.score_spans(
        examples=examples_spacy,
        attr="ents",
        allow_overlap=True,
        getter=no_misc_getter,
    )

    pos_scores = Scorer.score_token_attr(examples, "tag")
    pos_scores = Scorer.score_token_attr(examples, "tag")



"""This is a utility script for applying the augmentation to the DaNE test set
and including the metrics in the performance scores."""

import json
import sys

sys.path.append("../..")  # import dacy

import spacy
from spacy.scorer import Scorer
from spacy.training.augment import create_lower_casing_augmenter, dont_augment
from wasabi import msg

from dacy.augmenters import (
    create_keyboard_augmenter,
    create_pers_augmenter,
    create_spacing_augmenter,
    create_æøå_augmenter,
)
from dacy.datasets import dane, danish_names, female_names, male_names, muslim_names
from dacy.score import n_sents_score, score

dk_name_dict = danish_names()
muslim_name_dict = muslim_names()
f_name_dict = female_names()
m_name_dict = male_names()

dk_aug = create_pers_augmenter(
    dk_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
muslim_aug = create_pers_augmenter(
    muslim_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
f_aug = create_pers_augmenter(
    f_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
m_aug = create_pers_augmenter(
    m_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
punct_aug = create_pers_augmenter(
    muslim_name_dict,
    force_pattern_size=False,
    keep_name=True,
    patterns=["abbpunct"],
)

keyboard_aug_02 = create_keyboard_augmenter(
    doc_level=1,
    char_level=0.02,
    keyboard="QWERTY_DA",
)
keyboard_aug_05 = create_keyboard_augmenter(
    doc_level=1,
    char_level=0.05,
    keyboard="QWERTY_DA",
)
keyboard_aug_15 = create_keyboard_augmenter(
    doc_level=1,
    char_level=0.15,
    keyboard="QWERTY_DA",
)


# Change æ=ae, ø=oe, å=aa
æøå_aug = create_æøå_augmenter(doc_level=1, char_level=1)

# lower case text
lower_case_aug = create_lower_casing_augmenter(level=1)

# spacing
spacing_aug_05 = create_spacing_augmenter(doc_level=1, spacing_level=0.05)
spacing_aug = create_spacing_augmenter(doc_level=1, spacing_level=1)


# augmenter, Name, iterations, Description
n = 20
bootstrap_text = f" As this agmentation is stochastic it is repeated {n} times to obtain a consistent estimate and the mean is provided with its standard deviation in parenthesis."
augmenters = [
    (
        dont_augment,
        "No augmentation",
        1,
        "Applies no augmentation to the DaNE test set.",
    ),
    (
        æøå_aug,
        "Æøå Augmentation",
        1,
        "This augmentation replace the æ,ø, and å with their spelling variations ae, oe and aa respectively.",
    ),
    (lower_case_aug, "Lowercase", 1, "This augmentation lowercases all text."),
    (
        spacing_aug,
        "No Spacing",
        1,
        "This augmentation removed all spacing from the text.",
    ),
    (
        punct_aug,
        "Abbreviated first names",
        1,
        "This agmentation abbreviates the first names of entities. For instance 'Kenneth Enevoldsen' would turn to 'K. Enevoldsen'.",
    ),
    (
        keyboard_aug_02,
        "Keystroke errors 2%",
        n,
        "This agmentation simulate keystroke errors by replacing 2% of keys with a neighbouring key on a Danish QWERTY keyboard."
        + bootstrap_text,
    ),
    (
        keyboard_aug_05,
        "Keystroke errors 5%",
        n,
        "This agmentation simulate keystroke errors by replacing 5% of keys with a neighbouring key on a Danish QWERTY keyboard."
        + bootstrap_text,
    ),
    (
        keyboard_aug_15,
        "Keystroke errors 15%",
        n,
        "This agmentation simulate keystroke errors by replacing 15% of keys with a neighbouring key on a Danish QWERTY keyboard."
        + bootstrap_text,
    ),
    (
        dk_aug,
        "Danish names",
        n,
        "This agmentation replace all names with Danish names derived from Danmarks Statistik (2021)."
        + bootstrap_text,
    ),
    (
        muslim_aug,
        "Muslim names",
        n,
        "This agmentation replace all names with Muslim names derived from  Meldgaard (2005)."
        + bootstrap_text,
    ),
    (
        f_aug,
        "Female names",
        n,
        "This agmentation replace all names with Danish female names derived from Danmarks Statistik (2021)."
        + bootstrap_text,
    ),
    (
        m_aug,
        "Male names",
        n,
        "This agmentation replace all names with Danish male names derived from Danmarks Statistik (2021)."
        + bootstrap_text,
    ),
    (
        spacing_aug_05,
        "Spacing Augmention 5%",
        n,
        "This agmentation replace all names with Danish male names derived from Danmarks Statistik (2021)."
        + bootstrap_text,
    ),
]


def main(model, output):
    test = dane(splits=["test"])
    spacy.prefer_gpu()
    nlp = spacy.load(model)

    scorer = Scorer(nlp)
    scores = {}
    for aug, nam, k, desc in augmenters:
        msg.info(f"Running augmenter: {nam}")

        scores_ = score(
            corpus=test,
            apply_fn=nlp,
            augmenters=aug,
            score_fn=[scorer.score],
            k=k,
            nlp=nlp,
        )
        m_d = scores_[scores_.columns].mean().to_dict()
        s_d = scores_[scores_.columns].std().to_dict()
        m_d.pop("k")
        s_d.pop("k")
        scores[nam] = {"mean": m_d, "std": s_d, "k": k}

    for n in [5, 10]:
        nam = f"Input size augmentation {n} sentences"
        msg.info(f"Running augmenter: {nam}")
        scores_ = n_sents_score(
            n_sents=n,
            apply_fn=nlp,
            score_fn=[scorer.score],
            nlp=nlp,
        )
        m_d = scores_[scores_.columns].mean().to_dict()
        s_d = scores_[scores_.columns].std().to_dict()
        m_d.pop("k")
        s_d.pop("k")
        scores[nam] = {"mean": m_d, "std": s_d, "k": 1, "desc": desc}

    with open(output, "w") as f:
        json.dump(scores, f)
    msg.good(f"Finished. Evaluation metrics saved to '{output}'")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model",
        type=str,
        help="the model to evaluate",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=str,
        help="the json output filename",
        required=True,
    )

    args = parser.parse_args()
    main(args.model, args.output)

    # import os
    # os.chdir("training/v0.1.0")

    # main("training/small/dane/model-best", "metrics/yolo.json")



import spacy

import dacy  # noqa


def test_add_subjectivity():
    nlp = spacy.blank("da")
    nlp.add_pipe("dacy/subjectivity")
    texts = [
        "Analysen viser, at økonomien bliver forfærdelig dårlig",
        "Jeg tror alligevel, det bliver godt",
        "",
    ]
    actual = ["objective", "subjective", None]
    docs = list(nlp.pipe(texts))
    for d, a in zip(docs, actual):
        if a is None:
            assert d._.subjectivity is None
        else:
            assert d._.subjectivity == a


def test_add_berttone_polarity():
    nlp = spacy.blank("da")
    nlp.add_pipe("dacy/polarity")

    texts = [
        "Analysen viser, at økonomien bliver forfærdelig dårlig",
        "Jeg tror alligevel, det bliver godt",
    ]
    docs = list(nlp.pipe(texts))

    # text 0 should be more negative that text 1
    assert docs[0]._.polarity_prob["prob"][0] < docs[1]._.polarity_prob["prob"][0]


def test_add_bertemotion_laden():
    nlp = spacy.blank("da")
    nlp.add_pipe("dacy/emotionally_laden")
    doc = nlp("Der er et træ i haven")
    assert doc._.emotionally_laden == "no emotion"


def test_add_bertemotion_emo():
    nlp = spacy.blank("da")
    nlp.add_pipe("dacy/emotionally_laden")
    nlp.add_pipe("dacy/emotion")
    doc = nlp("Har i set at Tesla har landet en raket på månen? Det er vildt!!")
    assert doc._.emotion == "overasket/målløs"



import spacy

import dacy  # noqa


def test_add_hate_speech_detection():
    nlp = spacy.blank("da")
    nlp.add_pipe("dacy/hatespeech_detection")
    texts = ["senile gamle idiot", "hej har du haft en god dag"]
    actual = ["offensive", "not offensive"]
    docs = list(nlp.pipe(texts))
    for d, a in zip(docs, actual):
        if a is None:
            assert d._.is_offensive is None
        else:
            assert d._.is_offensive == a


def test_add_bertemotion_emo():
    nlp = spacy.blank("da")
    nlp.add_pipe("dacy/hatespeech_detection")
    nlp.add_pipe("dacy/hatespeech_classification")
    doc = nlp("senile gamle idiot")
    assert doc._.hate_speech_type == "sprogbrug"



from pathlib import Path

import pandas as pd
import spacy
import spacy_stanza
import stanza
from spacy.training.augment import create_lower_casing_augmenter, dont_augment

import dacy
from dacy.augmenters import (
    create_keyboard_augmenter,
    create_pers_augmenter,
    create_spacing_augmenter,
    create_æøå_augmenter,
)
from dacy.datasets import dane, danish_names, female_names, male_names, muslim_names
from dacy.score import n_sents_score, score

from .apply_fns.apply_fn_danlp import apply_danlp_bert
from .apply_fns.apply_fn_flair import apply_flair
from .apply_fns.apply_fn_nerda import apply_nerda
from .apply_fns.apply_fn_polyglot import apply_polyglot

# # to download the danlp and nerda you will have to set up a certificate:
# import ssl

# ssl._create_default_https_context = ssl._create_unverified_context

# Dataset
test = dane(splits=["test"])

# Augmenters - Create a list of augmenters we wish to apply to our model.

# randomly augment names
dk_name_dict = danish_names()
muslim_name_dict = muslim_names()
f_name_dict = female_names()
m_name_dict = male_names()

dk_aug = create_pers_augmenter(
    dk_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
muslim_aug = create_pers_augmenter(
    muslim_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
f_aug = create_pers_augmenter(
    f_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
m_aug = create_pers_augmenter(
    m_name_dict,
    force_pattern_size=True,
    keep_name=False,
    patterns=["fn", "fn,ln", "fn,ln,ln"],
)
punct_aug = create_pers_augmenter(
    muslim_name_dict,
    force_pattern_size=False,
    keep_name=True,
    patterns=["abbpunct"],
)


# randomly change 5%/15% of characters to a neighbouring key
keyboard_aug_02 = create_keyboard_augmenter(
    doc_level=1,
    char_level=0.02,
    keyboard="QWERTY_DA",
)
keyboard_aug_05 = create_keyboard_augmenter(
    doc_level=1,
    char_level=0.05,
    keyboard="QWERTY_DA",
)
keyboard_aug_15 = create_keyboard_augmenter(
    doc_level=1,
    char_level=0.15,
    keyboard="QWERTY_DA",
)

# Change æ=ae, ø=oe, å=aa
æøå_aug = create_æøå_augmenter(doc_level=1, char_level=1)

# lower case text
lower_case_aug = create_lower_casing_augmenter(level=1)

# spacing
spacing_aug_05 = create_spacing_augmenter(doc_level=1, spacing_level=0.05)
spacing_aug = create_spacing_augmenter(doc_level=1, spacing_level=1)

n = 20
# augmenter   name               n rep
augmenters = [
    (dont_augment, "No augmentation", 1),
    (keyboard_aug_02, "Keystroke errors 2%", n),
    (keyboard_aug_05, "Keystroke errors 5%", n),
    (keyboard_aug_15, "Keystroke errors 15%", n),
    (æøå_aug, "Æøå Augmentation", 1),
    (lower_case_aug, "Lowercase", 1),
    (dk_aug, "Danish names", n),
    (muslim_aug, "Muslim names", n),
    (f_aug, "Female names", n),
    (m_aug, "Male names", n),
    (punct_aug, "Abbreviated first names", 1),
    (spacing_aug_05, "Spacing Augmention 5%", n),
    (spacing_aug, "No Spacing", 1),
]

# Apply functions and models
# Loading application functions for necessary models. No need to create one for SpaCy pipelines.

model_dict = {
    "stanza": "da",
    "spacy_small": "da_core_news_sm",
    "spacy_medium": "da_core_news_md",
    "spacy_large": "da_core_news_lg",
    "dacy_small": "da_dacy_small_tft-0.0.0",
    "dacy_medium": "da_dacy_medium_tft-0.0.0",
    "dacy_large": "da_dacy_large_tft-0.0.0",
    "flair": apply_flair,
    "polyglot": apply_polyglot,
    "danlp_bert": apply_danlp_bert,
    "nerda_bert": apply_nerda,
}

# # Performance

Path("robustness").mkdir(parents=True, exist_ok=True)

for mdl in model_dict:
    print(f"[INFO]: Scoring model '{mdl}' using DaCy")

    # load model
    if "dacy" in mdl:
        apply_fn = dacy.load(model_dict[mdl])
    elif "spacy" in mdl:
        apply_fn = spacy.load(model_dict[mdl])
        spacy.prefer_gpu()
    elif "stanza" in mdl:
        stanza.download(model_dict[mdl])
        # Initialize the pipeline
        apply_fn = spacy_stanza.load_pipeline(model_dict[mdl])
    else:
        apply_fn = model_dict[mdl]

    i = 0
    scores = []
    for aug, nam, k in augmenters:
        print(f"\t Running augmenter: {nam}")

        scores_ = score(corpus=test, apply_fn=apply_fn, augmenters=aug, k=k)
        scores_["model"] = mdl
        scores_["augmenter"] = nam
        scores_["i"] = i
        scores.append(scores_)

        i += 1

    for n in [5, 10]:
        scores_ = n_sents_score(n_sents=n, apply_fn=apply_fn)
        scores_["model"] = mdl
        scores_["augmenter"] = f"Input size augmentation {n} sentences"
        scores_["i"] = i + 1
        scores.append(scores_)

    scores = pd.concat(scores)

    scores.to_csv(f"robustness/{mdl}_augmentation_performance.csv")



from .dane import dane  # noqa
from .names import danish_names  # noqa
from .names import female_names  # noqa
from .names import load_names  # noqa
from .names import male_names  # noqa
from .names import muslim_names  # noqa



from spacy.lang.da import Danish
from spacy.training import Corpus, Example

import dacy
from dacy.datasets import dane, female_names, male_names, muslim_names

test = dane(splits=["test"])


def test_dane():
    train, dev, test = dane(open_unverified_connection=True)
    all_ = dacy.datasets.dane(splits=["all"])
    for d in [train, dev, test, all_]:
        assert isinstance(d, Corpus)

    nlp = Danish()
    examples = list(test(nlp))  # check if it read as intended
    assert isinstance(examples[0], Example)

    all_ = dacy.datasets.dane(splits=["all"])


def test_names():
    for names in [muslim_names(), female_names(), male_names()]:
        assert isinstance(names["first_name"], list)
        assert len(names["first_name"]) > 0
        assert isinstance(names["last_name"], list)
        assert len(names["last_name"]) > 0



"""Helper functions for loading name dictionaries for person augmentation."""

import os
from typing import Dict, List, Optional

import pandas as pd


def load_names(
    min_count: int = 0,
    ethnicity: Optional[str] = None,
    gender: Optional[str] = None,
    min_prop_gender: float = 0,
) -> Dict[str, List[str]]:
    """Loads the names lookup table. Danish are from Danmarks statistik (2021).
    Muslim names are from Meldgaard (2005),
    https://nors.ku.dk/publikationer/webpublikationer/muslimske_fornavne/.

    Args:
        min_count (int, optional): Minimum number of occurences of the name for it to
            be included. Defaults to 0.
        ethnicity (Optional[str], optional): Which ethnicity should be included. None
            indicate all is included. Options include "muslim", "danish". Defaults to
            None.
        gender (Optional[str], optional): Which gender should be included. None indicate
            all is included. Options include "male", "female". Defaults to None.
        min_prop_gender (float): minimum probability of a name being a given gender.
            The probability of a given name being a specific gender is based on the
            proportion of people with the given name of that gender. Only used when
            gender is set. Defaults to 0.

    Returns:
        Dict[str, List[str]]: A dictionary of Muslim names containing the keys
            "first_name" and "last_name".
    """
    path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "lookup_tables",
        "names.csv",
    )
    names = pd.read_csv(path)

    if min_count:
        names = names.loc[names["count"] >= min_count]

    if ethnicity is not None:
        names = names.loc[names["ethnicity"] == ethnicity]

    last_names = names.loc[names["first_name"] == False]  # noqa
    if gender is not None:
        names = names.groupby(["name", "gender", "first_name"]).agg({"count": "sum"})
        # Change: groupby state_office and divide by sum
        names = names.groupby(level=0).apply(lambda x: x / float(x.sum()))
        names = names.reset_index()
        names = names.loc[
            (names["gender"] == gender) & (names["count"] >= min_prop_gender)
        ]

    first_names = names.loc[names["first_name"] == True]  # noqa
    return {
        "first_name": first_names.name.tolist(),
        "last_name": last_names.name.tolist(),
    }


def muslim_names() -> Dict[str, List[str]]:
    """Returns a dictionary of Muslim names.

    Returns:
        Dict[str, List[str]]: A dictionary of Muslim names containing the keys
            "first_name" and "last_name". The list is derived from Meldgaard (2005),
            https://nors.ku.dk/publikationer/webpublikationer/muslimske_fornavne/.

    Example:
        >>> from dacy.datasets import muslim_names
        >>> names = muslim_names()
        >>> names["first_name"]
        >>> names["last_name"]
    """
    return load_names(ethnicity="muslim")


def danish_names() -> Dict[str, List[str]]:
    """Returns a dictionary of Danish names.

    Returns:
        Dict[str, List[str]]: A dictionary of Danish names containing the keys
        "first_name" and "last_name". The list is derived from Danmarks statistik
        (2021).

    Example:
        >>> from dacy.datasets import danish_names
        >>> names = danish_names()
        >>> names["first_name"]
        >>> names["last_name"]
    """
    return load_names(ethnicity="danish")


def female_names() -> Dict[str, List[str]]:
    """Returns a dictionary of Danish female names.

    Returns:
        Dict[str, List[str]]: A dictionary of names containing the keys "first_name"
            and "last_name". The list is derived from Danmarks statistik (2021).

    Example:
        >>> from dacy.datasets import female_names
        >>> names = female_names()
        >>> names["first_name"]
        >>> names["last_name"]
    """
    return load_names(ethnicity="danish", gender="female", min_prop_gender=0.5)


def male_names() -> Dict[str, List[str]]:
    """Returns a dictionary of Danish male names.

    Returns:
        Dict[str, List[str]]: A dictionary of names containing the keys "first_name"
            and "last_name". The list is derived from Danmarks statistik (2021).

    Example:
        >>> from dacy.datasets import male_names
        >>> names = male_names()
        >>> names["first_name"]
        >>> names["last_name"]
    """
    return load_names(ethnicity="danish", gender="male", min_prop_gender=0.5)



from typing import Callable, List, Optional
from warnings import warn

from spacy.lang.da import Danish
from spacy.language import Language
from spacy.tokens import Doc
from spacy_transformers.data_classes import FullTransformerBatch
from spacy_wrap import (
    SequenceClassificationTransformer,
    make_sequence_classification_transformer,
)
from thinc.api import Config, Model

DEFAULT_CONFIG_STR = """
[subjectivity]
max_batch_items = 4096
doc_extension_trf_data = "subjectivity_trf_data"
doc_extension_prediction = "subjectivity"
labels = ["objective", "subjective"]

[subjectivity.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[subjectivity.model]
@architectures = "spacy-wrap.SequenceClassificationTransformerModel.v1"
name = "alexandrainst/da-subjectivivity-classification-base"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}

[subjectivity.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96


[polarity]
max_batch_items = 4096
doc_extension_trf_data = "polarity_trf_data"
doc_extension_prediction = "polarity"
labels =["positive", "neutral", "negative"]

[polarity.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[polarity.model]
@architectures = "spacy-wrap.SequenceClassificationTransformerModel.v1"
name = "alexandrainst/da-sentiment-base"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}

[polarity.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96


[emotionally_laden]
max_batch_items = 4096
doc_extension_trf_data = "emotionally_laden_trf_data"
doc_extension_prediction = "emotionally_laden"
labels = ["emotional", "no emotion"]

[emotionally_laden.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[emotionally_laden.model]
@architectures = "spacy-wrap.SequenceClassificationTransformerModel.v1"
name = "alexandrainst/da-binary-emotion-classification-base"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}

[emotionally_laden.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96


[emotion]
max_batch_items = 4096
doc_extension_trf_data = "emotion_trf_data"
doc_extension_prediction = "emotion"
labels = ["glæde/sindsro", "tillid/accept", "forventning/interrese", "overasket/målløs", "vrede/irritation", "foragt/modvilje", "sorg/trist", "frygt/bekymret"]


[emotion.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[emotion.model]
@architectures = "spacy-wrap.SequenceClassificationTransformerModel.v1"
name = "alexandrainst/da-emotion-classification-base"
tokenizer_config = {"use_fast": true}
transformer_config = {}
mixed_precision = false
grad_scaler_config = {}

[emotion.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96
"""  # noqa: E501

DEFAULT_CONFIG = Config().from_str(DEFAULT_CONFIG_STR)


Danish.factory(
    "dacy/subjectivity",
    default_config=DEFAULT_CONFIG["subjectivity"],
)(make_sequence_classification_transformer)


Danish.factory(
    "dacy/polarity",
    default_config=DEFAULT_CONFIG["polarity"],
)(make_sequence_classification_transformer)

Danish.factory(
    "dacy/emotionally_laden",
    default_config=DEFAULT_CONFIG["emotionally_laden"],
)(make_sequence_classification_transformer)


@Danish.factory(
    "dacy/emotion",
    default_config=DEFAULT_CONFIG["emotion"],
)
def make_emotion_transformer(
    nlp: Language,
    name: str,
    model: Model[List[Doc], FullTransformerBatch],
    set_extra_annotations: Callable[[List[Doc], FullTransformerBatch], None],
    max_batch_items: int,
    doc_extension_trf_data: str,
    doc_extension_prediction: str,
    labels: List[str],
) -> SequenceClassificationTransformer:
    if not Doc.has_extension("dacy/emotionally_laden"):
        warn(
            "The 'emotion' component assumes the 'emotionally_laden' extension is set."
            + " To set it you can run  nlp.add_pipe('dacy/emotionally_laden')",
        )

    # TODO: Add a conditional forward such that the model isn't run is document is not
    # emotionally laden
    clf_mdl = SequenceClassificationTransformer(
        vocab=nlp.vocab,
        model=model,
        set_extra_annotations=set_extra_annotations,
        max_batch_items=max_batch_items,
        name=name,
        labels=labels,
        doc_extension_trf_data=doc_extension_trf_data,
        doc_extension_prediction=doc_extension_prediction,
        assign_to_cats=True,
    )

    # overwrite extension such that it return no emotion if the document does not have
    # an emotion
    if Doc.has_extension("dacy/emotionally_laden"):

        def label_getter(doc) -> Optional[str]:
            if doc._.emotionally_laden == "emotional":
                prob = getattr(doc._, f"{doc_extension_prediction}_prob")
                if prob["prob"] is not None:
                    return labels[int(prob["prob"].argmax())]
            return doc._.emotionally_laden

        Doc.set_extension(doc_extension_prediction, getter=label_getter, force=True)
    clf_mdl.model.initialize()
    return clf_mdl



"""This is a utility script for updating the spacy meta.json.

Sample call
python --meta meta.json --augment metrics/dane_augmented_best_dacy_small_trf-0.1.0.json --
"""

import json


def create_description():
    from augment import augmenters

    describtion = """
<details>

<summary> Description of Augmenters </summary>

    """
    describtion
    for aug, nam, k, desc in augmenters:
        describtion += f"\n\n**{nam}:**\n{desc}"
    describtion += "\n </details> \n <br /> \n"
    return describtion


def main(meta_json, meta_augment_json, size, decimals=3):
    with open(meta_json) as f:
        meta = json.load(f)
    with open(meta_augment_json) as f:
        meta_augment = json.load(f)

    meta["email"] = "Kenneth.enevoldsen@cas.au.dk"
    meta["author"] = "Centre for Humanities Computing Aarhus"
    meta["url"] = "https://chcaa.io/#/"
    meta["license"] = "Apache-2.0 License"

    mdl_used = {
        "small": {
            "name": "Maltehb/-l-ctra-danish-electra-small-cased",
            "author": "Malte Højmark-Bertelsen",
            "url": "https://huggingface.co/Maltehb/-l-ctra-danish-electra-small-cased",
            "license": "CC BY 4.0",
        },
        "medium": {
            "name": "Maltehb/danish-bert-botxo",
            "author": "BotXO.ai",
            "url": "https://huggingface.co/Maltehb/danish-bert-botxo",
            "license": "CC BY 4.0",
        },
        "large": {
            "name": "xlm-roberta-large",
            "author": "Alexis Conneau, Kartikay Khandelwal, Naman Goyal, Vishrav Chaudhary, Guillaume Wenzek, Francisco Guzmán, Edouard Grave, Myle Ott, Luke Zettlemoyer, Veselin Stoyanov",
            "url": "https://huggingface.co/xlm-roberta-large",
            "license": "CC BY 4.0",
        },
    }
    model = mdl_used[size]

    meta["sources"] = [
        {
            "name": "UD Danish DDT v2.5",
            "url": "https://github.com/UniversalDependencies/UD_Danish-DDT",
            "license": "CC BY-SA 4.0",
            "author": "Johannsen, Anders; Mart\u00ednez Alonso, H\u00e9ctor; Plank, Barbara",
        },
        {
            "name": "DaNE",
            "url": "https://github.com/alexandrainst/danlp/blob/master/docs/datasets.md#danish-dependency-treebank-dane",
            "license": "CC BY-SA 4.0",
            "author": "Rasmus Hvingelby, Amalie B. Pauli, Maria Barrett, Christina Rosted, Lasse M. Lidegaard, Anders S\u00f8gaard",
        },
        model,
    ]

    meta["requirements"] = ["spacy-transformers>=1.0.3,<1.1.0"]

    meta[
        "description"
    ] = f"""
<a href="https://github.com/centre-for-humanities-computing/Dacy"><img src="https://centre-for-humanities-computing.github.io/DaCy/_static/icon.png" width="175" height="175" align="right" /></a>

# DaCy {size} transformer

DaCy is a Danish language processing framework with state-of-the-art pipelines as well as functionality for analysing Danish pipelines.
DaCy's largest pipeline has achieved State-of-the-Art performance on Named entity recognition, part-of-speech tagging and dependency 
parsing for Danish on the DaNE dataset. Check out the [DaCy repository](https://github.com/centre-for-humanities-computing/DaCy) for material on how to use DaCy and reproduce the results. 
DaCy also contains guides on usage of the package as well as behavioural test for biases and robustness of Danish NLP pipelines.
    """

    meta[
        "notes"
    ] = """
## Bias and Robustness

Besides the validation done by SpaCy on the DaNE testset, DaCy also provides a series of augmentations to the DaNE test set to see how well the models deal with these types of augmentations.
The can be seen as behavioural probes akinn to the NLP checklist.

### Deterministic Augmentations
Deterministic augmentations are augmentation which always yield the same result.

| Augmentation | Part-of-speech tagging (Accuracy) | Morphological tagging (Accuracy) | Dependency Parsing (UAS) | Dependency Parsing (LAS) | Sentence segmentation (F1) | Lemmatization (Accuracy) | Named entity recognition (F1) |
| --- | --- | --- | --- |  --- | --- | --- |  --- |
"""

    for aug, metrics in meta_augment.items():
        if metrics["k"] == 1:
            pos = f'{round(metrics["mean"]["pos_acc"], decimals)}'
            morph = f'{round(metrics["mean"]["morph_acc"], decimals)}'
            dep_uas = f'{round(metrics["mean"]["dep_uas"], decimals)}'
            dep_las = f'{round(metrics["mean"]["dep_las"], decimals)}'
            sent_f = f'{round(metrics["mean"]["sents_f"], decimals)}'
            lemma = f'{round(metrics["mean"]["lemma_acc"], decimals)}'
            ents_f = f'{round(metrics["mean"]["ents_f"], decimals)}'
            meta[
                "notes"
            ] += f"| {aug} | {pos} | {morph} | {dep_uas} |  {dep_las} | {sent_f} | {lemma} | {ents_f} |\n"

    meta[
        "notes"
    ] += """


### Stochastic Augmentations
Stochastic augmentations are augmentation which are repeated mulitple times to estimate the effect of the augmentation.

| Augmentation | Part-of-speech tagging (Accuracy) | Morphological tagging (Accuracy) | Dependency Parsing (UAS) | Dependency Parsing (LAS) | Sentence segmentation (F1) | Lemmatization (Accuracy) | Named entity recognition (F1) |
| --- | --- | --- | --- |  --- | --- | --- |  --- |
"""

    for aug, metrics in meta_augment.items():
        if metrics["k"] > 1:
            pos = f'{round(metrics["mean"]["pos_acc"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            morph = f'{round(metrics["mean"]["morph_acc"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            dep_uas = f'{round(metrics["mean"]["dep_uas"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            dep_las = f'{round(metrics["mean"]["dep_las"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            sent_f = f'{round(metrics["mean"]["sents_f"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            lemma = f'{round(metrics["mean"]["lemma_acc"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            ents_f = f'{round(metrics["mean"]["ents_f"], decimals)} ({round(metrics["std"]["pos_acc"], decimals)})'
            meta[
                "notes"
            ] += f"| {aug} | {pos} | {morph} | {dep_uas} |  {dep_las} | {sent_f} | {lemma} | {ents_f} |\n"

    meta["notes"] += create_description()

    meta[
        "notes"
    ] += "\n\n### Hardware\nThis was run an trained on a Quadro RTX 8000 GPU."

    with open(f"template_meta_{size}.json", "w") as f:
        json.dump(meta, f)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--meta",
        type=str,
        help="the meta file you wish to update",
        required=True,
    )
    parser.add_argument(
        "--augment",
        type=str,
        help="the json file of the augmented resutls",
        required=True,
    )
    parser.add_argument("--size", type=str, help="the model size", required=True)

    args = parser.parse_args()
    main(args.meta, args.augment, args.size)



def fetch_dane_as_conllu():
    from danlp.datasets import DDT

    ddt = DDT()
    train, dev, test = ddt.load_as_conllu(predefined_splits=True)
    with open("assets/dane/dane_train.conllu", "w") as f:
        train.write(f)
    with open("assets/dane/dane_dev.conllu", "w") as f:
        dev.write(f)
    with open("assets/dane/dane_test.conllu", "w") as f:
        test.write(f)


if __name__ == "__main__":
    fetch_dane_as_conllu()



from dacy.hate_speech import make_offensive_transformer  # noqa
from dacy.sentiment import make_emotion_transformer  # noqa

from .about import __download_url__, __title__, __version__  # noqa
from .download import download_model  # noqa
from .load import load, models, where_is_my_dacy  # noqa



"""Functionality for loading and locating DaCy models."""
from typing import Optional

from spacy.language import Language
from wasabi import msg

from .download import DEFAULT_CACHE_DIR, download_model, models_url


def load(
    model: str,
    path: Optional[str] = None,
    force_download: bool = False,
    **kwargs,
) -> Language:
    """Load a DaCy model as a SpaCy text processing pipeline. If the model is
    not downloaded it will also download the model.

    Args:
        model (str): the model you wish to load. To see available model see
            dacy.models()
        path (str, optional): The path to the downloaded model. Defaults to None which
            corresponds to the path optained using dacy.where_is_my_dacy().
        force_redownload (bool, optional): Should the model be redownloaded even if
            already downloaded? Default to False.
        kwargs: additional arguments passed to spacy.load()

    Returns:
        Language: a SpaCy text-preprocessing pipeline

    Example:
        >>> import dacy
        >>> dacy.load("da_dacy_medium_trf-0.1.0")
        >>> # or equivalently
        >>> dacy.load("medium")
    """
    import spacy

    if path is None:
        path = DEFAULT_CACHE_DIR

    path = download_model(model, path, force=force_download)
    return spacy.load(path, **kwargs)


def where_is_my_dacy(verbose: bool = True) -> str:
    """Returns a path to where DaCy models are located. The default the model
    location can be configured with the environmental variable
    `DACY_CACHE_DIR`.

    Args:
        verbose (bool, optional): Toggles the verbosity of the function. Defaults to
            True.

    Returns:
        str: path to the location of DaCy models

    Example:
        >>> import dacy
        >>> dacy.where_is_my_dacy()
    """
    if verbose is True:
        msg.info(
            "DaCy pipelines above and including version 0.1.0 are installed as a "
            + "python module and are thus located in your python environment under "
            + "the respective names. To get a list of installed models use "
            + "spacy.util.get_installed_models()",
        )
    return DEFAULT_CACHE_DIR


def models() -> list:
    """Returns a list of valid DaCy models.

    Returns:
        list: list of valid DaCy models

    Example:
        >>> import dacy
        >>> dacy.models()
    """
    return list(models_url.keys())



### pip install polyglot morfessor==2.0.6 pycld2==0.41 pyicu
### polyglot download pos2.da

from polyglot.tag import NEChunker, POSTagger
from polyglot.text import Text, WordList
from spacy.lang.da import Danish
from spacy.tokens import Doc
from spacy.training import Example

from .apply_fn_utils import add_iob, apply_on_multiple_examples, no_misc_getter

ne_chunker = NEChunker(lang="da")
pos_tagger = POSTagger(lang="da")


nlp_da = Danish()


def __apply_polyglot(example: Example, use_spacy: bool = False) -> Example:
    # tokenization
    if use_spacy:
        words = WordList(
            [t.text for t in nlp_da(example.reference.text)],
            language="da",
        )
    else:
        text = Text(example.reference.text, hint_language_code="da")
        words = text.words
    # ner
    iob = [iob for token, iob in ne_chunker.annotate(words)]
    # pos-tagging
    tags = [tag for t, tag in pos_tagger.annotate(words)]

    words = [t.string for t in words]
    doc = Doc(nlp_da.vocab, words=words, tags=tags, pos=tags)
    doc = add_iob(doc, iob)
    return Example(doc, example.y)


apply_polyglot = apply_on_multiple_examples(__apply_polyglot)

if __name__ == "__main__":
    import os

    os.chdir("..")
    from dacy.datasets import dane

    test = dane(splits=["test"])
    nlp = Danish()
    examples = apply_polyglot(test(nlp), use_spacy=False)
    examples_spacy = apply_polyglot(test(nlp), use_spacy=True)

    from spacy.scorer import Scorer

    tok_scores = Scorer.score_tokenization(examples)
    ent_scores = Scorer.score_spans(
        examples=examples,
        attr="ents",
        getter=no_misc_getter,
    )
    ent_scores_spacy = Scorer.score_spans(
        examples=examples_spacy,
        attr="ents",
        getter=no_misc_getter,
    )
    pos_scores = Scorer.score_token_attr(examples, "tag")

    from spacy import displacy

    displacy.render(examples[0].y, style="ent")
    displacy.render(examples[0].x, style="ent")



from collections.abc import MutableMapping


def flatten_dict(d: dict, sep: str = "_") -> dict:
    """flattens a nested dictionary.

    Args:
        d (dict): The nested dictionary to be flattened. Keys are assumed to be strings.
        sep (str, optional): The seperator between keys. Defaults to '_'.

    Returns:
        dict: the flattened dict.

    Example:
        >>> flatten({'a': 1, 'c': {'a': 2, 'b': {'x': 5, 'y' : 10}}, 'd': [1, 2, 3]})
        {'a': 1, 'c_a': 2, 'c_b_x': 5, 'd': [1, 2, 3], 'c_b_y': 10}
    """

    def __inner_flatten(d, parent_key, sep):
        items = []
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, MutableMapping):
                items.extend(__inner_flatten(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    return __inner_flatten(d, "", sep)



# to download the danlp and nerda you will have to set up a certificate:
import ssl
from typing import Iterable, List

from NERDA.precooked import DA_BERT_ML
from spacy.lang.da import Danish
from spacy.tokens import Doc
from spacy.training import Example

from .apply_fn_utils import add_iob, no_misc_getter

ssl._create_default_https_context = ssl._create_unverified_context

model = DA_BERT_ML()
# model.download_network()
model.load_network()
nlp_da = Danish()


def apply_nerda(examples: Iterable[Example], use_spacy: bool = True) -> List[Example]:
    sentences = []
    docs_y = []
    for example in examples:
        # tokenization
        if use_spacy:
            sentences.append([t.text for t in nlp_da(example.reference.text)])
        else:
            from nltk.tokenize import word_tokenize

            sentences.append(word_tokenize(example.reference.text))
        docs_y.append(example.reference)

    # ner
    labels = model.predict(sentences=sentences)

    examples_ = []
    for doc_y, label, words in zip(docs_y, labels, sentences):
        if len(label) < len(words):
            label += ["O"] * (len(words) - len(label))

        doc = Doc(nlp_da.vocab, words=words)
        doc = add_iob(doc, iob=label)
        examples_.append(Example(doc, doc_y))
    return examples_


if __name__ == "__main__":
    import os

    os.chdir("..")
    from dacy.datasets import dane

    test = dane(splits=["test"])
    nlp = Danish()
    examples = apply_nerda(test(nlp), use_spacy=True)

    from spacy.scorer import Scorer

    tok_scores = Scorer.score_tokenization(examples)
    ent_scores = Scorer.score_spans(
        examples=examples,
        attr="ents",
        getter=no_misc_getter,
    )
    pos_scores = Scorer.score_token_attr(examples, "tag")

    from spacy import displacy

    displacy.render(examples[0].y, style="ent")
    displacy.render(examples[0].x, style="ent")



import spacy


def test_ner():
    nlp = spacy.blank("da")
    # or nlp = dacy.load("da_dacy_small_tft-0.0.0", exclude=["ner"])
    nlp.add_pipe("dacy/ner")

    doc = nlp("Jeg hedder Peter og bor i København")
    assert doc.ents[0].text == "Peter"
    assert doc.ents[0].label_ == "PER"
    assert doc.ents[1].text == "København"
    assert doc.ents[1].label_ == "LOC"



"""This includes function for scoring models applied to a SpaCy corpus."""
from __future__ import annotations

from copy import copy
from functools import partial
from time import time
from typing import Callable, Iterable

import pandas as pd
from spacy.language import Language
from spacy.scorer import Scorer
from spacy.tokens import Doc, Span
from spacy.training import Corpus, Example, dont_augment

from ..utils import flatten_dict


def no_misc_getter(doc: Doc, attr: str) -> Iterable[Span]:
    """A utility getter for scoring entities without including MISC.

    Args:
        doc (Doc): a SpaCy Doc
        attr (str): attribute to be extracted

    Returns:
        Iterable[Span]
    """
    spans = getattr(doc, attr)
    for span in spans:
        if span.label_ == "MISC":
            continue
        yield span


def dep_getter(token, attr):
    dep = getattr(token, attr)
    dep = token.vocab.strings.as_string(dep).lower()
    return dep


def score(  # noqa
    corpus: Corpus,
    apply_fn: Callable[[Iterable[Example], list[Example]]] | Language,
    score_fn: list[Callable[[Iterable[Example]], dict] | str] = [
        "token",
        "pos",
        "ents",
        "dep",
    ],
    augmenters: list[Callable[[Language, Example], Iterable[Example]]] = [],
    k: int = 1,
    nlp: Language | None = None,
    **kwargs,
) -> pd.DataFrame:
    """scores a models performance on a given corpus with potentially
    augmentations applied to it.

    Args:
        corpus (Corpus): A spacy Corpus
        apply_fn (Union[Callable, Language]): A wrapper function for the model you wish
            to score. The model should take in a list of spacy Examples
            (Iterable[Example]) and output a tagged version of it (Iterable[Example]). A
            SpaCy pipeline (Language) can be provided as is.
        score_fn (List[Union[Callable[[Iterable[Example]], dict], str]], optional): A
            scoring function which takes in a list of examples (Iterable[Example]) and
            return a dictionary of performance scores. Four potiential strings are
            valid. "ents" for measuring the performance of entity spans. "pos" for
            measuring the performance of fine-grained (tag_acc), and coarse-grained
            (pos_acc) pos-tags. "token" for measuring the performance of tokenization.
            "dep" for measuring the performance of dependency parsing. "nlp" for
            measuring the performance of all components in the specified nlp pipeline.
            Defaults to ["token", "pos", "ents", "dep"].
        augmenters (List[Callable[[Language, Example], Iterable[Example]]], optional): A
            spaCy style augmenters which should be applied to the corpus or a list
            thereof. defaults to [], indicating no augmenters.
        k (int, optional): Number of times it should run the augmentation and test the
            performance on the corpus. Defaults to 1.
        nlp (Optional[Language], optional): A spacy processing pipeline. If None it will
            use an empty Danish pipeline. Defaults to None. Used for loading the calling
            the corpus.

    Returns:
        pandas.DataFrame: returns a pandas dataframe containing the performance metrics.

    Example:
        >>> from spacy.training.augment import create_lower_casing_augmenter
        >>> from dacy.datasets import dane
        >>> test = dane(splits=["test")
        >>> nlp = dacy.load("da_dacy_small_tft-0.0.0")
        >>> scores = score(test, augmenter=[create_lower_casing_augmenter(0.5)],
        >>>                apply_fn = nlp)
    """
    if callable(augmenters):
        augmenters = [augmenters]
    if len(augmenters) == 0:
        augmenters = [dont_augment]

    def __apply_nlp(examples):
        examples = ((e.x.text, e.y) for e in examples)
        doc_tuples = nlp_.pipe(examples, as_tuples=True)
        return [Example(x, y) for x, y in doc_tuples]

    if isinstance(apply_fn, Language):
        nlp_ = apply_fn
        apply_fn = __apply_nlp

    if nlp is None:
        from spacy.lang.da import Danish

        nlp = Danish()

    scorer = Scorer(nlp)

    def ents_scorer(examples):
        scores = Scorer.score_spans(examples, attr="ents")
        scores_no_misc = Scorer.score_spans(
            examples,
            attr="ents",
            getter=no_misc_getter,
        )
        scores["ents_excl_MISC"] = {
            k: scores_no_misc[k] for k in ["ents_p", "ents_r", "ents_f"]
        }
        return scores

    def pos_scorer(examples):
        scores = Scorer.score_token_attr(examples, attr="pos")
        scores_ = Scorer.score_token_attr(examples, attr="tag")
        for k in scores_:
            scores[k] = scores_[k]
        return scores

    def_scorers = {
        "ents": ents_scorer,
        "pos": pos_scorer,
        "token": Scorer.score_tokenization,
        "nlp": scorer.score,
        "dep": partial(
            Scorer.score_deps,
            attr="dep",
            getter=dep_getter,
            ignore_labels=("p", "punct"),
        ),
    }

    def __score(augmenter):
        corpus_ = copy(corpus)
        corpus_.augmenter = augmenter
        scores_ls = []
        for i in range(k):
            s = time()
            examples = apply_fn(corpus_(nlp))
            speed = time() - s
            scores = {"wall_time": speed}
            for fn in score_fn:
                if isinstance(fn, str):
                    fn = def_scorers[fn]
                scores.update(fn(examples))
            scores = flatten_dict(scores)
            scores_ls.append(scores)

        # and collapse list to dict
        for key in scores.keys():
            scores[key] = [s[key] if key in s else None for s in scores_ls]

        scores["k"] = list(range(k))

        return pd.DataFrame(scores)

    for i, aug in enumerate(augmenters):
        scores_ = __score(aug)
        scores = pd.concat([scores, scores_]) if i != 0 else scores_  # noqa
    return scores



"""This includes the DaNE dataset wrapped and read in as a SpaCy corpus."""

import os
import shutil
from pathlib import Path
from typing import List, Optional, Union

from spacy.training import Corpus

from ..download import DEFAULT_CACHE_DIR, download_url
from .constants import DATASETS


def dane(
    save_path: Optional[str] = None,
    splits: List[str] = ["train", "dev", "test"],
    redownload: bool = False,
    n_sents: int = 1,
    open_unverified_connection: bool = False,
    **kwargs,
) -> Union[List[Corpus], Corpus]:
    """Reads the DaNE dataset as a spacy Corpus.

    Args:
        save_path (str, optional): Path to the DaNE dataset If it does not contain the
            dataset it is downloaded to the folder. Defaults to None corresponding to
            dacy.where_is_my_dacy() in the datasets subfolder.
        splits (List[str], optional): Which splits of the dataset should be returned.
            Possible options include "train", "dev", "test", "all". Defaults to
            ["train", "dev", "test"].
        redownload (bool, optional): Should the dataset be redownloaded. Defaults to
            False.
        n_sents (int, optional): Number of sentences per document. Only applied if the
            dataset is downloaded. Defaults to 1.
        open_unverified_connection (bool, optional): Should you download from an
            unverified connection. Defaults to False.
        force_extension (bool, optional): Set the extension to the doc regardless of
            whether it already exists. Defaults to False.

    Returns:
        Union[List[Corpus], Corpus]: Returns a SpaCy corpus or a list thereof.

    Example:
        >>> from dacy.datasets import dane
        >>> train, dev, test = dane(splits=["train", "dev", "test"])
    """
    if open_unverified_connection:
        import ssl

        ssl._create_default_https_context = ssl._create_unverified_context

    if save_path is None:
        save_path_ = os.path.join(DEFAULT_CACHE_DIR, "datasets")
    else:
        save_path_ = save_path
    save_path = os.path.join(save_path_, "dane")

    if (
        (not os.path.isdir(save_path))
        or ("dane" not in os.listdir(save_path_))
        or (redownload is True)
    ):
        Path(save_path).mkdir(parents=True, exist_ok=True)

        dl_path = os.path.join(save_path, "dane.zip")
        download_url(DATASETS["dane"], dl_path)
        shutil.unpack_archive(dl_path, save_path)
        os.remove(dl_path)

    wpaths = [
        "dane_train.conllu",
        "dane_dev.conllu",
        "dane_test.conllu",
        "dane.conllu",
    ]

    for wpath in wpaths:
        wpath = os.path.join(save_path, wpath)
        cpath = wpath[:-7] + f"_{n_sents}"
        if os.path.isfile(cpath + ".spacy"):
            continue
        cpath += ".conllu"
        shutil.copyfile(wpath, cpath)
        # convert to spacy
        os.system(
            f"python -m spacy convert {cpath} {save_path} --converter conllu "
            + f"--merge-subtokens -n {n_sents}",
        )
        os.remove(cpath)

    if isinstance(splits, str):
        splits = [splits]
    corpora = []
    paths = {
        "all": f"dane_{n_sents}.spacy",
        "test": f"dane_test_{n_sents}.spacy",
        "dev": f"dane_dev_{n_sents}.spacy",
        "train": f"dane_train_{n_sents}.spacy",
    }

    for split in splits:
        corpora.append(Corpus(os.path.join(save_path, paths[split])))
    if len(corpora) == 1:
        return corpora[0]
    return corpora
